from __future__ import absolute_import, division, print_function, unicode_literals
# int is not imported from builtins here for performance reasons.
# See: https://github.com/PythonCharmers/python-future/issues/136
from builtins import (ascii, bytes, chr, dict, filter, hex, input,
                      map, next, oct, open, pow, range, round,
                      str, super, zip, object)

import forgi.threedee.utilities.vector as ftuv
import forgi.threedee.utilities.graph_pdb as ftug
import forgi.utilities.stuff as fus
import collections as col
import numpy as np
import itertools as it
# import networkx as nx Takes to long. Import only when needed
import warnings
import math
import copy
import sys
import logging


log = logging.getLogger(__name__)

"""
This module uses code by David Eppstein
(http://code.activestate.com/recipes/117225-convex-hull-and-diameter-of-2d-point-sets/)
under the PSF license and code by Syrtis Major
(c)2014-2015 (http://stackoverflow.com/a/24567352/5069869) under the BSD 3-Clause license
and potentially copyrighted code from matplotlib under the PSF license
(http://matplotlib.org/examples/api/line_with_text.html).
"""

try:
    profile  # The @profile decorator from line_profiler (kernprof)
except NameError:
    def profile(x):
        return x


def to_rgb(im):
    """
    Convert an np array image from grayscale to RGB
    """
    # SEE http://www.socouldanyone.com/2013/03/converting-grayscale-to-rgb-with-numpy.html
    try:
        w, h = im.shape
    except ValueError:  # Probably RGB already
        w, h, c = im.shape
        return im
    else:
        newImg = np.empty((w, h, 3), dtype=np.uint8)
        newImg[:, :, 2] = newImg[:, :, 1] = newImg[:, :, 0] = (
            im * 255).astype(np.uint8)
        return newImg


def to_grayscale(im):
    """
    Convert an np array image from RGB to grayscale
    """
    try:
        w, h, _ = im.shape
    except ValueError:
        return im
    else:
        newImg = np.empty((w, h))
        newImg[:, :] = (im[:, :, 0] / 255 + im[:, :, 1] /
                        255 + im[:, :, 2] / 255) / 3
        return newImg


def rasterized_2d_coordinates(points, angstrom_per_cell=10, origin=np.array([0, 0]), rotate=0):
    angle = math.radians(rotate)
    c = np.cos(angle)
    s = np.sin(angle)
    rotMat = np.array([[c, -s], [s, c]])
    #print("FPP: rasterization: rotationmatrix ",rotMat)
    a = ((np.dot(points, rotMat) - origin) // angstrom_per_cell).astype(int)
    b = np.dot(points, rotMat)
    #print("FPP: rasterization: roatated: ", b)
    b = b - origin
    #print("FPP: rasterization: roatated and offset: ", b)
    assert (a == (b // angstrom_per_cell).astype(int)).all()
    return a


def crop_coordinates_to_bounds(a, num_cells):
    """
    :a: an array of x,y coordinates (modified in place)
    """
    return a.clip(min=0, max=num_cells - 1)

# The following functions are from
# http://code.activestate.com/recipes/117225-convex-hull-and-diameter-of-2d-point-sets/
# Used under the PSF License
############################################################################################
# convex hull (Graham scan by x-coordinate) and diameter of a set of points
# David Eppstein, UC Irvine, 7 Mar 2002


def orientation(p, q, r):
    '''Return positive if p-q-r are clockwise, neg if ccw, zero if colinear.'''
    return (q[1] - p[1]) * (r[0] - p[0]) - (q[0] - p[0]) * (r[1] - p[1])

#@profile


def hulls(Points):
    '''Graham scan to find upper and lower convex hulls of a set of 2d points.'''
    U = []
    L = []
    Points.sort(key=lambda x: (x[0], x[1]))
    for p in Points:
        while len(U) > 1 and orientation(U[-2], U[-1], p) <= 0:
            U.pop()
        while len(L) > 1 and orientation(L[-2], L[-1], p) >= 0:
            L.pop()
        U.append(p)
        L.append(p)
    return U, L


def rotatingCalipers(Points):
    '''Given a list of 2d points, finds all ways of sandwiching the points
between two parallel lines that touch one point each, and yields the sequence
of pairs of points touched by each pair of lines.'''
    U, L = hulls(Points)
    i = 0
    j = len(L) - 1
    while i < len(U) - 1 or j > 0:
        yield U[i], L[j]

        # if all the way through one side of hull, advance the other side
        if i == len(U) - 1:
            j -= 1
        elif j == 0:
            i += 1

        # still points left on both lists, compare slopes of next hull edges
        # being careful to avoid divide-by-zero in slope calculation
        elif (U[i + 1][1] - U[i][1]) * (L[j][0] - L[j - 1][0]) > \
                (L[j][1] - L[j - 1][1]) * (U[i + 1][0] - U[i][0]):
            i += 1
        else:
            j -= 1


def diameter(Points):
    '''Given a list of 2d points, returns the pair that's farthest apart.'''
    try:
        diam, pair = max([((p[0] - q[0])**2 + (p[1] - q[1])**2, (p, q))
                          for p, q in rotatingCalipers(Points)], key=lambda x: x[0])
    except:
        print(repr([((p[0] - q[0])**2 + (p[1] - q[1])**2, (p, q))
                    for p, q in rotatingCalipers(Points)]))
        raise
    return pair


# END David Eppstein

#@profile
def rotate2D(vector, cosPhi, sinPhi):
    x = vector[0] * cosPhi - vector[1] * sinPhi
    y = vector[0] * sinPhi + vector[1] * cosPhi
    return np.array([x, y])


def bresenham(start, end):
    """
    Rasterize a line from start to end onto a grid with grid-width 1.

    :param start: A sequence of length 2, containing the x,y coordinates of the start of the line
    :param start: A sequence of length 2, containing the x,y coordinates of the end of the line
    :returns: A list of tuples (x,y), where x and y are integers.
    """
    # See e.g. http://stackoverflow.com/a/32252934/5069869
    # or https://de.wikipedia.org/wiki/Bresenham-Algorithmus#C-Implementierung
    if start == end:
        return [start]
    points = []
    dx = end[0] - start[0]
    dy = end[1] - start[1]
    x, y = start
    if dx == 0:
        sx = 0
    else:
        sx = dx // abs(dx)
    if dy == 0:
        sy = 0
    else:
        sy = dy // abs(dy)
    dx = abs(dx)
    dy = abs(dy)
    if dx > dy:
        err = dx / 2.
        while x != end[0]:
            # print(x,y)
            points.append((x, y))
            err -= dy
            if err < 0:
                y += sy
                err += dx
            x += sx
    else:
        err = dy / 2.
        while y != end[1]:
            # print(x,y)
            points.append((x, y))
            err -= dx
            if err < 0:
                x += sx
                err += dy
            y += sy
    points.append((x, y))
    # if abs(dx)>1 or abs(dy)>1:
    #print(start, end, points)
    return points


class Projection2D(object):
    """
    A 2D Projection of a CoarseGrainRNA unto a 2D-plane
    """
    @profile
    def __init__(self, cg, proj_direction=None, rotation=0, project_virtual_atoms=False,
                 project_virtual_residues=[]):
        """
        :param cg: a CoarseGrainRNA object with 3D coordinates for every element

                   .. note::
                      The projection is generated from this cg, but it is not associated
                      with it after construction.
                      Thus future changes of the cg are not reflected in the projection.

        :param proj_direction: a carthesian vector (in 3D space) in the direction of projection.
                               The length of this vector is not used.
                               If proj_direction is None, cg.project_from is used.
                               If proj_direction and cg.project_from is None, an error is raised.
        :param rotate: Degrees. Rotate the projection by this amount.

        """
        #: The projected coordinates of all stems
        self._coords = dict()

        self._cross_points = None
        self._proj_graph = None

        # Calculate orthonormal basis of projection plane.
        # Compare to none, because `if np.array:` raises ValueError.
        if proj_direction is not None:
            proj_direction = np.array(proj_direction, dtype=float)
        elif cg.project_from is not None:
            # We make a copy here. In case cg.project_from is modified,
            # we still want to be able to look up from what direction the projection was generated.
            proj_direction = np.array(cg.project_from, dtype=float)
        else:
            raise ValueError(
                "No projection direction given and none present in the cg Object.")
        _, unit_vec1, unit_vec2 = ftuv.create_orthonormal_basis(proj_direction)
        self._unit_vec1 = unit_vec1
        self._unit_vec2 = unit_vec2
        self._proj_direction = proj_direction
        self._virtual_residues = []
        self.virtual_residue_numbers = project_virtual_residues
        self._project(cg, project_virtual_atoms, project_virtual_residues)

        # Rotate and translate projection into a standard orientation
        points = list(self.points)
        v1, v2 = diameter(points)
        #: The longest distance between any two points of the projection.
        self.longest_axis = ftuv.vec_distance(v1, v2)

        v1 = np.array(v1)
        v2 = np.array(v2)
        shift = (v1 + v2) / 2
        for key, edge in self._coords.items():
            self._coords[key] = (edge[0] - shift, edge[1] - shift)

        if project_virtual_atoms:
            self._virtual_atoms = self._virtual_atoms - shift
        if project_virtual_residues:
            self._virtual_residues = self._virtual_residues - shift
        rot = math.atan2(*(v2 - v1))
        rot = math.degrees(rot)
        self.rotate(rot)
        xmean = np.mean([x[0] for p in self._coords.values() for x in p])
        ymean = np.mean([x[1] for p in self._coords.values() for x in p])
        mean = np.array([xmean, ymean])
        for key, edge in self._coords.items():
            self._coords[key] = (edge[0] - mean, edge[1] - mean)
        # Thanks to numpy broadcasting, this works without a loop.
        if project_virtual_atoms:
            self._virtual_atoms = self._virtual_atoms - mean
        if project_virtual_residues:
            self._virtual_residues = self._virtual_residues - mean
        # From this, further rotate if requested by the user.
        if rotation != 0:
            self.rotate(rotation)

    ### Functions modifying the projection in place ###
    @profile
    def rotate(self, angle):
        """
        Rotate the projection in place around the origin (0,0)

        :param angle: Rotation angle in degrees.
        """
        angle = math.radians(angle)
        c = np.cos(angle)
        s = np.sin(angle)
        self._proj_graph = None
        for key, edge in self._coords.items():
            self._coords[key] = (rotate2D(edge[0], c, s),
                                 rotate2D(edge[1], c, s))

        transRotMat = np.array([[c, s], [-s, c]])
        if len(self._virtual_atoms):
            self._virtual_atoms = np.dot(self._virtual_atoms, transRotMat)
        if self.virtual_residue_numbers:
            self._virtual_residues = np.dot(
                self._virtual_residues, transRotMat)

    def condense_points(self, cutoff=1):
        """
        Condenses several projection points that are within a range of less than cutoff into
        one point. This function modifies this Projection2D object.

        .. note::
           The result depends on the ordering of the dictionary holding
           the nodes and might thus be pseudorandomly

        :param cutoff: Two point with a distance smaller than cuttoff are contracted.
                       A value below 20 is reasonable.
        """
        if cutoff <= 0:
            return
        while self._condense_one(cutoff):
            pass
        self.proj_graph.remove_edges_from(self.proj_graph.selfloop_edges())

    def condense(self, cutoff):
        """
        Condenses points that are within the cutoff of another point or edge (=line segment)
        into a new point. This function modifies this Projection2D object.

        The contraction of two points works like documented in
        `self.condense_points(self, cutoff)`.
        If a node is close to a line segment, a new point is generated between
        this node and the line segment. Then the original line and the original node
        are deleted and all connections attached to the new point.

        .. note::
           The result depends on the ordering of the dictionary holding the nodes and
           might thus be pseudorandomly


        :param cutoff: Two point with a distance smaller than cuttoff are contracted.
                       A value below 20 is reasonable.
        """
        if cutoff <= 0:
            return
        self.condense_points(cutoff)
        while self._condense_pointWithLine_step(cutoff):
            self.condense_points(cutoff)

    ### Get virtual residues ###
    @property
    def vres_iterator(self):
        for i, nr in enumerate(self.virtual_residue_numbers):
            yield nr, self._virtual_residues[i]

    def get_vres_by_position(self, pos):
        """
        :param pos: The nucleotide position in the sequence
        """
        for i, nr in enumerate(self.virtual_residue_numbers):
            if nr == pos:
                return self._virtual_residues[i]
        else:
            raise LookupError("This virtual residue was not projected")
    ### Properties ###

    @property
    def proj_direction(self):
        """A vector describing the direction of the projection"""
        return self._proj_direction

    @property
    def crossingPoints(self):
        """
        All points, where 2 segments intersect

        A list of triples `(key1, key2, coordinate)`, where coordinate is a 2D vector.
        """
        if self._cross_points is None:
            self._cross_points = col.defaultdict(list)
            for key1, key2 in it.combinations(self._coords, 2):
                for cr in ftuv.seg_intersect(self._coords[key1], self._coords[key2]):
                    self._cross_points[key1].append((cr, key2))
                    self._cross_points[key2].append((cr, key1))
        return self._cross_points

    # Note: This is SLOW the first time it is called. Should be avoided as much as possible.
    @property
    def proj_graph(self):
        """A graph describing the projected RNA.

        This graph is stored as a `networkx` graph object.
        """
        if self._proj_graph is None:
            self._build_proj_graph()
        return self._proj_graph

    @property
    def points(self):
        """
        All points that are at the ends of coarse-grain elements.
        This does not include points where coarse grain elements intersect in the projection.

        Some points might be duplicates

        :returns: A generator yielding all points.
        """

        # Exclude m and i Elements, as they are always flanked by stems.
        return (p for k, x in self._coords.items() for p in x if k[0] != "i" and k[0] != "m")
    ### Functions returning descriptors of the projection, mostly independent on the resolution ###

    ### Function returning descriptors of the projection, dependent on the resolution ###
    def get_bounding_box(self, margin=0.):
        """
        Returns the coordinates for a box that contains all points of the 2D projection.

        :param margin: increase the bounding box in every direction by this margin.
        :returns: left, right, bottom, top
        """
        points = [p for x in self._coords.values() for p in x]
        #print("P", points)
        left = min(x[0] for x in points) - margin
        right = max(x[0] for x in points) + margin
        bottom = min(x[1] for x in points) - margin
        top = max(x[1] for x in points) + margin
        # print "BB",  left, right, bottom, top
        return left, right, bottom, top

    def get_bounding_square(self, margin=0.):
        """
        Returns the coordinates for a square that contains all points of the 2D projection.

        :param margin: increase the bounding box in every direction by this margin.
        :returns: left, right, bottom, top
        """
        bb = self.get_bounding_box()
        length = max([bb[1] - bb[0], bb[3] - bb[2]]) / 2 + margin
        x = (bb[0] + bb[1]) / 2
        y = (bb[2] + bb[3]) / 2
        return x - length, x + length, y - length, y + length

    def get_branchpoint_count(self, degree=None):
        """
        Returns the number of branchpoint.

        .. note::
           This measure is sensitive to the resolution of a projection.
           In an AFM image, one might not see all branching points.

        .. warning::
            Whether this code will stay in the library or not depends
            on future evaluation of the usefulness of this and similar descriptors.

        :param degree: If degree is None, count all points with degree>=3
                       Else: only count (branch)points of the given degree
        """
        import networkx as nx
        assert int(nx.__version__[
                   0]) >= 2, "This function only works with networkx version 2, not {}!".format(nx.__version__)
        if degree is None:
            return len([x[1] for x in nx.degree(self.proj_graph) if x[1] >= 3])
        else:
            return len([x[1] for x in nx.degree(self.proj_graph) if x[1] == degree])

    def get_cyclebasis_len(self):
        """
        Returns the number of cycles of length>1 in the cycle basis.

        .. warning::
            Whether this code will stay in the library or not depends
            on future evaluation of the usefulness of this and similar descriptors.

        """
        import networkx as nx
        return len([x for x in nx.cycle_basis(self.proj_graph) if len(x) > 1])

    def get_total_length(self):
        """
        Returns the sum of the lengths of all edges in the projection graph.

        .. note::
           This measure is sensitive to the resolution of a projection.
           In an AFM image, one might not see all cycles.

        .. warning::
            Whether this code will stay in the library or not depends
            on future evaluation of the usefulness of this and similar descriptors.
        """
        l = 0
        for edge in self.proj_graph.edges():
            l += ftuv.vec_distance(edge[0], edge[1])
        return l

    def get_longest_arm_length(self):
        """
        Get the length of the longest arm.

        An arm is a simple path from a node of `degree!=2` to a node of `degree 1`,
        if all the other nodes on the path have `degree 2`.

        .. note:: This measure is sensitive to the resolution of a projection
                  the same way the length of a coastline is sensitive to the resolution.

        .. warning::
            Whether this code will stay in the library or not depends
            on future evaluation of the usefulness of this and similar descriptors.

        :returns: The length and a tuple of points `(leaf_node, corresponding_branch_point)`
        """
        import networkx as nx
        lengths = {}
        target = {}
        for leaf, degree in nx.degree(self.proj_graph):
            if degree != 1:
                continue
            lengths[leaf] = 0
            previous = None
            current = leaf
            while True:
                next = [x for x in self.proj_graph[current].keys() if x !=
                        previous]
                assert len(next) == 1
                next = next[0]
                lengths[leaf] += ftuv.vec_distance(current, next)
                if self.proj_graph.degree(next) != 2:
                    break
                previous = current
                current = next
            target[leaf] = next
        best_leaf = max(lengths, key=lambda x: lengths[x])
        return lengths[best_leaf], (best_leaf, target[best_leaf])

    def get_leaf_leaf_distances(self):
        """
        Get a list of distances between any pair of leaf nodes.
        The distances are measured in direct line, not along the path

        .. warning::
            Whether this code will stay in the library or not depends
            on future evaluation of the usefulness of this and similar descriptors.

        :returns: a list of floats (lengths in Angstrom)
        """
        lengths = []
        leaves = [leaf for leaf in self.proj_graph.nodes(
        ) if self.proj_graph.degree(leaf) == 1]
        for leaf1, leaf2 in it.combinations(leaves, 2):
            lengths.append(ftuv.vec_distance(leaf1, leaf2))
        lengths.sort(reverse=True)
        return lengths

    def get_some_leaf_leaf_distances(self):
        """
        Get a list of distances between some pairs of leaf nodes.
        The distances are measured in direct line, not along the path

        .. warning::
            Whether this code will stay in the library or not depends
            on future evaluation of the usefulness of this and similar descriptors.

        :returns: a list of floats (lengths in Angstrom)
        """
        lengths = []
        leaves = [leaf for leaf in self.proj_graph.nodes(
        ) if self.proj_graph.degree(leaf) == 1]
        for leaf1, leaf2 in it.combinations(leaves, 2):
            lengths.append((ftuv.vec_distance(leaf1, leaf2), leaf1, leaf2))
        lengths.sort(reverse=True, key=lambda x: x[0])
        newlengths = []
        visited = set()
        for l, leaf1, leaf2 in lengths:
            if leaf1 in visited or leaf2 in visited:
                continue
            newlengths.append(l)
            visited.add(leaf1)
            visited.add(leaf2)
        return newlengths

    def get_maximal_path_length(self):
        """
        Get the maximal path length from all simple paths that traverses the projection graph from
        one leave node to another.

        .. note::
           This measure is sensitive to the resolution of a projection
           the same way the length of a coastline is sensitive to the resoltuion.

        .. warning::
            Whether this code will stay in the library or not depends
            on future evaluation of the usefulness of this and similar descriptors.

        """
        import networkx as nx
        maxl = 0
        for i, node1 in enumerate(self.proj_graph.nodes()):
            for j, node2 in enumerate(self.proj_graph.nodes()):
                if j <= i:
                    continue
                all_paths = nx.all_simple_paths(self.proj_graph, node1, node2)
                for path in all_paths:
                    l = self._get_path_length(path)
                    if l > maxl:
                        maxl = l
        return maxl

    ### Functions for graphical representations of the projection ###
    @profile
    def rasterize(self, resolution=50, bounding_square=None, warn=True,
                  virtual_atoms=True, rotate=0, virtual_residues=True):
        """
        Rasterize the projection to a square image of the given resolution.
        Uses the Bresenham algorithm for line rasterization.

        :param resolution:
                        The number of pixels in each direction.

        :param bounding_square:
                        Rasterize onto the given square.
                        If `None`, automatically get a bounding_square that
                        shows the whole projection

        :param warn:    If True, raise a warning if parts of the projection are not inside
                        the given bounding square.

        :param virtual_atoms:
                        If True, virtual atoms are also rasterized.

        :param rot:     The in-plane rotation in degrees, applied before rotation.

        :returns:       A tuple `(np.array, float)`. The first value is a resolution x resolution
                        numpy 2D array.
                        The values are floats from 0.0 (black) to 1.0 (white).
                        This array can be directly plotted using matplotlib:
                        `pyplot.imshow(array, cmap='gray', interpolation='none')`
                        The second value is the length of one pixle in angstrom.
        """
        if bounding_square is None:
            bounding_square = self.get_bounding_square()
        box = bounding_square
        steplength = (box[1] - box[0]) / resolution
        image = np.zeros([resolution, resolution], dtype=np.float32)
        img_length = len(image)
        angle = math.radians(rotate)
        c = np.cos(angle)
        s = np.sin(angle)
        starts = []
        ends = []
        for label in self._coords:
            if virtual_atoms and len(self._virtual_atoms):
                if label[0] == "s":
                    continue
            starts.append(self._coords[label][0])
            ends.append(self._coords[label][1])
        starts = np.array(starts)
        ends = np.array(ends)
        starts = rasterized_2d_coordinates(
            starts, steplength, np.array([box[0], box[2]]), rotate)
        ends = rasterized_2d_coordinates(
            ends, steplength, np.array([box[0], box[2]]), rotate)
        for i in range(len(starts)):
            points = bresenham(tuple(starts[i]), tuple(ends[i]))
            for p in points:
                if 0 <= p[0] < img_length and 0 <= p[1] < img_length:
                    image[p[0], p[1]] = 1
                else:
                    if warn:
                        warnings.warn("WARNING during rasterization of the 2D Projection: "
                                      "Parts of the projection are cropped off.")
        if virtual_atoms and len(self._virtual_atoms):
            rot_virtual_atoms = rasterized_2d_coordinates(
                self._virtual_atoms, steplength, np.array([box[0], box[2]]), rotate)
            rot_virtual_atoms_clip = crop_coordinates_to_bounds(
                rot_virtual_atoms, img_length)
            if warn and (rot_virtual_atoms_clip != rot_virtual_atoms).any():
                warnings.warn("WARNING during rasterization of virtual atoms: "
                              "Parts of the projection are cropped off.")
            image[rot_virtual_atoms_clip[:, 0],
                  rot_virtual_atoms_clip[:, 1]] = 1
        if virtual_residues and self.virtual_residue_numbers:
            image = to_rgb(image)
            rot_virtual_res = rasterized_2d_coordinates(
                self._virtual_residues, steplength, np.array([box[0], box[2]]), rotate)
            for i, point in enumerate(rot_virtual_res):
                color = (
                    0, 150, 255 * self.virtual_residue_numbers[i] // max(self.virtual_residue_numbers))
                if 0 <= point[0] < img_length and 0 <= point[1] < img_length:
                    image[point[0], point[1]] = color
                else:
                    if warn:
                        warnings.warn("WARNING during rasterization of virtual residues: "
                                      "Parts of the projection are cropped off.")
        return image, steplength

    def plot(self, ax=None, show=False, margin=5,
             linewidth=None, add_labels=False,
             line2dproperties={}, xshift=0, yshift=0,
             show_distances=[], print_distances=False,
             virtual_atoms=True):
        """
        Plots the 2D projection.

        This uses modified copy-paste code by Syrtis Major (c)2014-2015
        under the BSD 3-Clause license and code from matplotlib under the PSF license.

        :param ax:          The axes to draw to.
                            You can get it by calling `fig, ax=matplotlib.pyplot.subplots()`

        :param show:        If true, the matplotlib.pyplot.show() will be called at the
                            end of this function.

        :param margin:      A numeric value.
                            The margin around the plotted projection inside the (sub-)plot.

        :param linewidth:   The width of the lines projection.

        :param add_labels:  Display the name of the corresponding coarse grain element in
                            the middle of each segment in the projection.
                            Either a bool or a set of labels to display.

        :param line2dproperties:
                            A dictionary. Will be passed as `**kwargs` to the constructor of
                            `matplotlib.lines.Line2D`.
                            See http://matplotlib.org/api/lines_api.html#matplotlib.lines.Line2D

        :param xshift, yshift:
                            Shift the projection by the given amount inside the canvas.

        :param show_distances:
                            A list of tuples of strings, e.g. `[("h1","h8"),("h2","m15")]`.
                            Show the distances between these elements in the plot

        :param print_distances:
                            Bool. Print all distances from show_distances at the side of the plot
                            instead of directly next to the distance
        """
        # In case of ssh without -X option, a TypeError might be raised
        # during the import of pyplot
        # This probably depends on the version of some library.
        # This is also the reason why we import matplotlib only inside the plot function.
        text = []
        try:
            if ax is None or show:
                import matplotlib.pyplot as plt
            import matplotlib.lines as lines
            import matplotlib.transforms as mtransforms
            import matplotlib.text as mtext
            import matplotlib.font_manager as font_manager
        except TypeError as e:
            warnings.warn("Cannot plot projection. Maybe you could not load Gtk "
                          "(no X11 server available)? During the import of matplotlib"
                          "the following Error occured:\n {}: {}".format(type(e).__name__, e))
            return
        except ImportError as e:
            warnings.warn("Cannot import matplotlib. Do you have matplotlib installed? "
                          "The following error occured:\n {}: {}".format(type(e).__name__, e))
            return
        # try:
        #    import shapely.geometry as sg
        #    import shapely.ops as so
        # except ImportError as e:
        #    warnings.warn("Cannot import shapely. "
        #                  "The following error occured:\n {}: {}".format(type(e).__name__, e))
        #    area=False
        #    #return
        # else:
        #    area=True
        area = False
        polygons = []

        def circles(x, y, s, c='b', ax=None, vmin=None, vmax=None, **kwargs):
            """
            Make a scatter of circles plot of x vs y, where x and y are sequence
            like objects of the same lengths. The size of circles are in data scale.

            Parameters
            ----------
            x,y : scalar or array_like, shape (n, )
                Input data
            s : scalar or array_like, shape (n, )
                Radius of circle in data scale (ie. in data unit)
            c : color or sequence of color, optional, default : 'b'
                `c` can be a single color format string, or a sequence of color
                specifications of length `N`, or a sequence of `N` numbers to be
                mapped to colors using the `cmap` and `norm` specified via kwargs.
                Note that `c` should not be a single numeric RGB or
                RGBA sequence because that is indistinguishable from an array of
                values to be colormapped.  `c` can be a 2-D array in which the
                rows are RGB or RGBA, however.
            ax : Axes object, optional, default: None
                Parent axes of the plot. It uses gca() if not specified.
            vmin, vmax : scalar, optional, default: None
                `vmin` and `vmax` are used in conjunction with `norm` to normalize
                luminance data.  If either are `None`, the min and max of the
                color array is used.  (Note if you pass a `norm` instance, your
                settings for `vmin` and `vmax` will be ignored.)

            Returns
            -------
            paths : `~matplotlib.collections.PathCollection`

            Other parameters
            ----------------
            kwargs : `~matplotlib.collections.Collection` properties
                eg. alpha, edgecolors, facecolors, linewidths, linestyles, norm, cmap

            Examples
            --------
            a = np.arange(11)
            circles(a, a, a*0.2, c=a, alpha=0.5, edgecolor='none')

            License
            --------
            This function is copied (and potentially modified) from
            http://stackoverflow.com/a/24567352/5069869

            Copyright Syrtis Major, 2014-2015

            This function is under [The BSD 3-Clause License]
            (http://opensource.org/licenses/BSD-3-Clause)
            """
            from matplotlib.patches import Circle
            from matplotlib.collections import PatchCollection
            #import matplotlib.colors as colors
            if ax is None:
                raise TypeError()

            if fus.is_string_type(c):
                color = c     # ie. use colors.colorConverter.to_rgba_array(c)
            else:
                color = None  # use cmap, norm after collection is created
            kwargs.update(color=color)

            if np.isscalar(x):
                patches = [Circle((x, y), s), ]
            elif np.isscalar(s):
                patches = [Circle((x_, y_), s) for x_, y_ in zip(x, y)]
            else:
                patches = [Circle((x_, y_), s_) for x_, y_, s_ in zip(x, y, s)]
            collection = PatchCollection(patches, **kwargs)

            if color is None:
                collection.set_array(np.asarray(c))
                if vmin is not None or vmax is not None:
                    collection.set_clim(vmin, vmax)

            ax.add_collection(collection)
            ax.autoscale_view()
            return collection

        class MyLine(lines.Line2D):
            """
            Copied and modified from http://matplotlib.org/examples/api/line_with_text.html,
            which is part of matplotlib 1.5.0 (Copyright (c) 2012-2013 Matplotlib Development
            Team; All Rights Reserved).
            Used under the matplotlib license: http://matplotlib.org/users/license.html
            """

            def __init__(self, *args, **kwargs):
                # we'll update the position when the line data is set
                fm = font_manager.FontProperties(size="large", weight="demi")
                self.text = mtext.Text(0, 0, '', fontproperties=fm)
                lines.Line2D.__init__(self, *args, **kwargs)

                # we can't access the label attr until *after* the line is
                # inited
                self.text.set_text(self.get_label())

            def set_figure(self, figure):
                self.text.set_figure(figure)
                lines.Line2D.set_figure(self, figure)

            def set_axes(self, axes):
                self.text.set_axes(axes)
                lines.Line2D.set_axes(self, axes)

            def set_transform(self, transform):
                # 2 pixel offset
                texttrans = transform + mtransforms.Affine2D().translate(2, 2)
                self.text.set_transform(texttrans)
                lines.Line2D.set_transform(self, transform)

            def set_data(self, x, y):
                if len(x):
                    self.text.set_position(
                        ((x[0] + x[-1]) / 2, (y[0] + y[-1]) / 2))
                lines.Line2D.set_data(self, x, y)

            def draw(self, renderer):
                # draw my label at the end of the line with 2 pixel offset
                lines.Line2D.draw(self, renderer)
                self.text.draw(renderer)

        if "linewidth" in line2dproperties and linewidth is not None:
            warnings.warn(
                "Got multiple values for 'linewidth' (also present in line2dproperties)")
        if linewidth is not None:
            line2dproperties["linewidth"] = linewidth
        if "solid_capstyle" not in line2dproperties:
            line2dproperties["solid_capstyle"] = "round"

        if ax is None:
            try:
                fig, ax = plt.subplots(1, 1)
            except Exception as e:
                warnings.warn("Cannot create Axes  or Figure. You probably have no graphical "
                              "display available. The Error was:\n {}: {}".format(type(e).__name__, e))
                return
        lprop = copy.copy(line2dproperties)
        if virtual_atoms and len(self._virtual_atoms) > 0:
            circles(
                self._virtual_atoms[:, 0], self._virtual_atoms[:, 1], c="gray", s=0.7, ax=ax)
        for label, (s, e) in self._coords.items():
            if "color" not in line2dproperties:
                if label.startswith("s"):
                    lprop["color"] = "green"
                elif label.startswith("i"):
                    lprop["color"] = "gold"
                elif label.startswith("h"):
                    lprop["color"] = "blue"
                elif label.startswith("m"):
                    lprop["color"] = "red"
                elif label.startswith("f") or label.startswith("t"):
                    lprop["color"] = "blue"
                else:
                    lprop["color"] = "black"
            if add_labels != False and (add_labels == True or label in add_labels):
                lprop["label"] = label
            else:
                lprop["label"] = ""
            #line=lines.Line2D([s[0], e[0]],[s[1],e[1]], **lprop)
            line = MyLine([s[0] + xshift, e[0] + xshift],
                          [s[1] + yshift, e[1] + yshift], **lprop)
            ax.add_line(line)
            s = s + np.array([xshift, yshift])
            e = e + np.array([xshift, yshift])
            vec = np.array(e) - np.array(s)
            nvec = np.array([vec[1], -vec[0]])
            try:
                div = math.sqrt(nvec[0]**2 + nvec[1]**2)
            except ZeroDivisionError:
                div = 100000
            a = e + nvec * 5 / div
            b = e - nvec * 5 / div
            c = s + nvec * 5 / div
            d = s - nvec * 5 / div
            # For now disabling area representation
            area = False
            if area:
                polygon = sg.Polygon([a, b, d, c])
                polygons.append(polygon)
        for s, e in show_distances:
            st = (self._coords[s][0] + self._coords[s][1]) / 2
            en = (self._coords[e][0] + self._coords[e][1]) / 2
            d = ftuv.vec_distance(st, en)
            if print_distances:
                line = MyLine([st[0] + xshift, en[0] + xshift], [st[1] + yshift, en[1] + yshift],
                              color="orange", linestyle="--")
                text.append("{:3} - {:3}: {:5.2f}".format(s, e, d))
            else:
                line = MyLine([st[0] + xshift, en[0] + xshift], [st[1] + yshift, en[1] + yshift],
                              label=str(round(d, 1)), color="orange", linestyle="--")
            ax.add_line(line)

        ax.axis(self.get_bounding_square(margin))
        fm = font_manager.FontProperties(["monospace"], size="x-small")
        if print_distances:
            ax.text(0.01, 0.05, "\n".join(["Distances:"] + text),
                    transform=ax.transAxes, fontproperties=fm)
        if area:
            rnaArea = so.cascaded_union(polygons)
            rnaXs, rnaYs = rnaArea.exterior.xy
            ax.fill(rnaXs, rnaYs, alpha=0.5)
        out = ax.plot()
        if show:
            plt.show()
            return
        return out

    ### Private functions ###
    # Note: This is SLOW. Should be improved or removed in the future
    def _build_proj_graph(self):
        """
        Generate a graph from the 2D projection.

        This is implemented as a networkx.Graph with the coordinates as nodes.
        """
        import networkx as nx
        proj_graph = nx.Graph()
        for key, element in self._coords.items():
            crs = self.crossingPoints
            sortedCrs = ftuv.sortAlongLine(element[0], element[1], [
                                           x[0] for x in crs[key]])
            oldpoint = None
            for point in sortedCrs:
                point = (point[0], point[1])  # Tuple, to be hashable
                if oldpoint is not None:
                    proj_graph.add_edge(
                        oldpoint, point, attr_dict={"label": key})
                oldpoint = point
        self._proj_graph = proj_graph
        self.condense_points(0.00000000001)  # To avoid floating point problems

    @profile
    def _project(self, cg, project_virtual_atoms, project_virtual_residues):
        """
        Calculates the 2D coordinates of all coarse grained elements by vector rejection.
        Stores them inside self._coords
        """
        self._coords = dict()
        self._virtual_atoms = []
        basis = np.array([self._unit_vec1, self._unit_vec2]).T
        # Project all coordinates to this plane
        for key in cg.sorted_element_iterator():
            val = cg.coords[key]
            start = np.dot(val[0], basis)
            end = np.dot(val[1], basis)
            self._coords[key] = (start, end)
        va = []
        if project_virtual_atoms:
            for residuePos in range(1, cg.seq_length + 1):
                residue = cg.virtual_atoms(residuePos)
                if project_virtual_atoms == "selected":
                    try:
                        va.append(residue['P'])
                    except KeyError:
                        pass
                    try:
                        va.append(residue["C1'"])
                    except KeyError:
                        assert False  # Should never happen
                    try:
                        va.append(residue['C1'])
                    except KeyError:
                        pass
                    try:
                        va.append(residue["O3'"])
                    except KeyError:
                        pass
                else:
                    for pos in residue.values():
                        va.append(pos)
            if va:
                self._virtual_atoms = np.dot(np.array(va), basis)
            else:
                warnings.warn("No virtual atoms present in {} of length {}!".format(
                    cg.name, len(cg.seq)))
        vr = []
        for res in project_virtual_residues:
            vr.append(np.dot(cg.get_virtual_residue(res, True), basis))
        assert len(vr) == len(project_virtual_residues)
        if vr:
            self._virtual_residues = np.array(vr)

    def _condense_one(self, cutoff):
        """
        Condenses two adjacent projection points into one.

        :returns: True if a condensation was done, False if no condenstaion is possible.
        """
        for i, node1 in enumerate(self.proj_graph.nodes()):
            for j, node2 in enumerate(self.proj_graph.nodes()):
                if j <= i:
                    continue
                if ftuv.vec_distance(node1, node2) < cutoff:
                    newnode = ftuv.middlepoint(node1, node2)
                    # self.proj_graph.add_node(newnode)
                    for neighbor in list(self.proj_graph.adj[node1].keys()):
                        self.proj_graph.add_edge(newnode, neighbor,
                                                 attr_dict=self.proj_graph.adj[node1][neighbor])
                    for neighbor in list(self.proj_graph.adj[node2].keys()):
                        self.proj_graph.add_edge(newnode, neighbor,
                                                 attr_dict=self.proj_graph.adj[node2][neighbor])
                    if newnode != node1:  # Equality can happen because of floating point inaccuracy
                        self.proj_graph.remove_node(node1)
                    if newnode != node2:
                        self.proj_graph.remove_node(node2)
                    return True
        return False

    def _condense_pointWithLine_step(self, cutoff):
        """
        Used by `self.condense(cutoff)` as a single condensation step of a point
        with a line segment.
        """
        for i, source in enumerate(self.proj_graph.nodes()):
            for j, target in enumerate(self.proj_graph.nodes()):
                if j > i and self.proj_graph.has_edge(source, target):
                    for k, node in enumerate(self.proj_graph.nodes()):
                        if k == i or k == j:
                            continue
                        nearest = ftuv.closest_point_on_seg(
                            source, target, node)
                        nearest = tuple(nearest)
                        if nearest == source or nearest == target:
                            continue
                        if (ftuv.vec_distance(nearest, node) < cutoff):
                            newnode = ftuv.middlepoint(node, tuple(nearest))
                            attr_dict = self.proj_graph.adj[source][target]
                            self.proj_graph.remove_edge(source, target)
                            if source != newnode:
                                self.proj_graph.add_edge(
                                    source, newnode, attr_dict=attr_dict)
                            if target != newnode:
                                self.proj_graph.add_edge(target, newnode,
                                                         attr_dict=attr_dict)
                            if newnode != node:  # Equality possible bcse of floating point inaccuracy
                                for neighbor in self.proj_graph.adj[node].keys():
                                    attr_dict = self.proj_graph.adj[node][neighbor]
                                    self.proj_graph.add_edge(newnode, neighbor,
                                                             attr_dict=attr_dict)
                                self.proj_graph.remove_node(node)
                            return True
        return False

    def _get_path_length(self, path):
        """
        :param path: a list of nodes
        """
        l = 0
        for i in range(len(path) - 1):
            l += ftuv.vec_distance(path[i], path[i + 1])
        return l



from __future__ import division
from builtins import object

import forgi.threedee.model.coarse_grain as ftmc
import forgi.utilities.debug as fud

import forgi.threedee.utilities.vector as ftuv
import forgi.threedee.utilities.graph_pdb as ftug

import itertools as it
import logging
import math
import numpy as np
from collections import defaultdict


log = logging.getLogger(__name__)
__all__ = ['AdjacencyCorrelation', 'cg_rmsd', 'rmsd', 'drmsd']

"""
This module contains functions for the comparison of two cg objects or two ordered point-clouds.
"""

try:
    profile
except:
    def profile(f):
        return f


class Incompareable(ValueError):
    """
    Raised if two objects are compared, which are incompareable.

    E.g. Coordinate sets of different length are incompareable with respect to the RMSD.
    """


def ppv(tp, fp):
    '''
    Calculate the positive predictive value.

    :param tp: The true positives
    :param fp: The false positives
    :return: The ppv
    '''
    try:
        return tp / float(tp + fp)
    except ZeroDivisionError:
        return float("nan")


def sty(tp, fn):
    '''
    Calculate the sensitivity.

    :param tp: True positives
    :param fn: False negatives
    :return: The sensitivity
    '''
    try:
        return tp / float(tp + fn)
    except ZeroDivisionError:
        return float("nan")


def mcc(confusion_matrix):
    '''
    Calculate the Matthews correlation coefficient for
    the given confusion matrix.

    `MCC = sqrt(ppv * sty)`

    :param confustion_matrix: A dictionary like this: `{"tp": tp, "tn": tn, "fp": fp, "fn": fn}`
    :return: The MCC
    '''

    return math.sqrt(ppv(confusion_matrix['tp'],
                         confusion_matrix['fp']) *
                     sty(confusion_matrix['tp'],
                         confusion_matrix['fn']))


class AdjacencyCorrelation(object):
    """
    A class used for calculating the ACC.

    The adjacency correlation coefficient is calculated as the Matthews correlation
    coefficient of potential interactions, defined as nucleotides within 25 Angstrom
    from each other. See chapter 9.3 of Peter's thesis.

    This object is initialized with a reference structure and a distance for interactions.
    The evaluate() method is used for calculating this correlation matrix.

    This is significantly faster than the confusion_matrix function, if
    many structures will be compared to the same reference structure.
    """

    def __init__(self, reference_cg, distance=25.0, bp_distance=16):
        self._distance = distance
        self._bp_distance = bp_distance
        self._reference_interactions = self._get_interactions(reference_cg)

    @profile
    def _get_interactions(self, cg):
        """
        :return: A set of 2-tuples containing elements that pair.
        """
        ignore = set()
        for n1, n2 in it.combinations(cg.defines.keys(), r=2):
            if cg.connected(n1, n2):
                ignore.add((n1, n2))
                continue
            bp_dist = cg.min_max_bp_distance(n1, n2)[0]
            if bp_dist < self._bp_distance:
                ignore.add((n1, n2))

        interactions = set(
            cg.coords.elements_closer_than(self._distance, ignore))
        return interactions

    def evaluate(self, cg):
        '''
        Calculate the true_positive, false_positive,
        true_negative and false_negative rate for the tertiary
        distances of the elements of cg and the reference structure stored in this class.

        :param cg: The coarse grain model with which self.reference_cg should be compared.
                   Note: The result is only meaningful, if both coarse grained models
                   correspond to the same RNA.
        :return: A dictionary like this: `{"tp": tp, "tn": tn, "fp": fp, "fn": fn}`
        '''
        interactions = self._get_interactions(cg)
        nodes = set(cg.defines.keys())
        allIA = set()
        for n1, n2 in it.combinations(nodes, r=2):
            if cg.connected(n1, n2):
                continue
            bp_dist = cg.min_max_bp_distance(n1, n2)[0]
            if bp_dist < self._bp_distance:
                continue
            allIA.add(tuple(sorted((n1, n2))))

        d = {"tp": 0, "tn": 0, "fp": 0, "fn": 0}
        d["tp"] = len(self._reference_interactions & interactions)
        d["fp"] = len(interactions - self._reference_interactions)
        d["fn"] = len(self._reference_interactions - interactions)
        d["tn"] = len(allIA - (self._reference_interactions | interactions))
        return d


def optimal_superposition(crds1, crds2):
    """
    Returns best-fit rotation matrix as [3x3] numpy matrix for aligning crds1 onto crds2
    using the Kabsch algorithm
    """
    if crds1.shape != crds2.shape:
        raise Incompareable(
            "Cannot superimpose coordinate lists of different length.")
    if crds1.shape[1] == 3 or crds1.shape[1] == 2:
        correlation_matrix = np.dot(np.transpose(crds1), crds2)
        v, s, w_tr = np.linalg.svd(correlation_matrix)
        is_reflection = (np.linalg.det(v) * np.linalg.det(w_tr)) < 0.0
        if is_reflection:
            v[:, -1] = -v[:, -1]
        return np.dot(v, w_tr)
    else:
        raise ValueError("Wrong dimension of crds1. Needs to be an array of "
                         "Points in 2D or 3D space. Found {}D".format(crds1.shape[1]))

def cg_stem_rmsd(cg1, cg2):
    coords1 = cg1.get_ordered_stem_poss()
    coords2 = cg2.get_ordered_stem_poss()
    return rmsd(coords1, coords2)


def cg_rmsd(cg1, cg2):
    '''
    Calculate the RMSD between two Coarse Grain models using their
    set of virtual residues.

    :param cg1: The first coarse grain model.
    :param cg2: The second coarse-grain model.
    :return: The RMSD
    '''

    residues1 = cg1.get_ordered_virtual_residue_poss()
    residues2 = cg2.get_ordered_virtual_residue_poss()
    try:
        # First, try all coordinates, no matter if the seq-ids match
        return rmsd(residues1, residues2)
    except Incompareable:
        # If the number of nts is not equal, issue a warning and try to match based on seq_ids.
        res_ids = set(cg1.seq._seqids) | set(cg2.seq._seqids)
        common_resids = set(cg1.seq._seqids) & set(cg2.seq._seqids)
        residues1 = []
        residues2 = []
        for res in common_resids:
            residues1.append(cg1.get_virtual_residue(res, allow_single_stranded = True))
            residues2.append(cg2.get_virtual_residue(res, allow_single_stranded = True))
        if len(residues1)>len(res_ids)*0.8:
            log.warning("Using only %s common residues for RMSD comparison based on seq_ids.", len(residues1))
            if set(cg1.seq._seqids)-common_resids :
                log.warning("Ignoring from cg1:  %s.", set(cg1.seq._seqids)-common_resids )
            if set(cg2.seq._seqids)-common_resids:
                log.warning("Ignoring from cg2:  %s.", set(cg2.seq._seqids)-common_resids )
            return rmsd(residues1, residues2)
        else:
            log.warning("Cannot compare based on seqids: Intersection of"
                        " resids is too small: %s", common_resids)
        raise Incompareable("Cgs {} and {} cannot be compared according to the RMSD, "
                            "because they do not have the same number of "
                            "virtual residues.".format(cg1.name, cg2.name))


def rmsd_contrib_per_element(cg1, cg2):
    residues1, elems1 = cg1.get_ordered_virtual_residue_poss(
        return_elements=True)
    residues2, elems2 = cg2.get_ordered_virtual_residue_poss(
        return_elements=True)
    if elems1 != elems2:
        raise ValueError(
            "RNAs with different structure are not compareable by RMSD.")
    diff_vecs = _pointwise_deviation(residues1, residues2)
    elem_devs = defaultdict(list)
    for i, elem in enumerate(elems1):
        elem_devs[elem].append(diff_vecs[i])
    return elem_devs


def _pointwise_deviation(crds1, crds2, is_centered=False):
    """
    Helper function for Kabsch RMSD
    """
    if not is_centered:
        crds1 = ftuv.center_on_centroid(crds1)
        crds2 = ftuv.center_on_centroid(crds2)

    os = optimal_superposition(crds1, crds2)
    crds_aligned = np.dot(crds1, os)

    diff_vecs = (crds2 - crds_aligned)

    return diff_vecs


def rmsd_kabsch(crds1, crds2, is_centered=False):
    '''
    Center the coordinate vectors on their centroid
    and then calculate the rmsd.
    '''
    diff_vecs = _pointwise_deviation(crds1, crds2, is_centered)

    vec_lengths = np.sum(diff_vecs * diff_vecs, axis=1)
    return math.sqrt(sum(vec_lengths) / len(vec_lengths))


def drmsd(coords1, coords2):
    '''
    Calculate the dRMSD measure.

    This should be the RMSD between all of the inter-atom distances
    in two structures.

    :param coords1: The vectors of the 'atoms' in the first structure.
    :param coords2: The vectors of the 'atoms' in the second structure.
    :return: The dRMSD measure.
    '''
    ds1 = np.array([ftuv.vec_distance(c1, c2)
                    for c1, c2 in it.combinations(coords1, r=2)])
    ds2 = np.array([ftuv.vec_distance(c1, c2)
                    for c1, c2 in it.combinations(coords2, r=2)])

    rmsd = math.sqrt(np.mean((ds1 - ds2) * (ds1 - ds2)))
    #rmsd = math.sqrt(np.mean(ftuv.vec_distance(ds1, ds2)))

    return rmsd


def rmsd_qc_wrap(coords1, coords2, is_centered=False):
    r = rmsd_qc(coords1, coords2, is_centered)
    if np.isnan(r):
        return rmsd_kabsch(coords1, coords2, is_centered)
    return r


try:
    from py_qcprot import rmsd as rmsd_qc  # Faster C version, if available
    rmsd = rmsd_qc_wrap
except:
    rmsd = rmsd_kabsch


def basepair_distance(cg1, cg2):
    # QUESTION Move to forgi.graph? or forgi.utilities
    # Note: An implementation in c (with python bindings) is available in the Vienna RNA package
    dist = 0
    if str(cg1.seq) != str(cg2.seq):  # Compare as strings, to ignore missing and mofified residues
        raise Incompareable(
            "We do not support a basepair distance between rnas with different sequences.")
    for stem in cg1.stem_iterator():
        for bp in cg1.stem_bp_iterator(stem):
            if cg2.pairing_partner(bp[0]) != bp[1]:
                dist += 1
    for stem in cg2.stem_iterator():
        for bp in cg2.stem_bp_iterator(stem):
            if cg1.pairing_partner(bp[0]) != bp[1]:
                dist += 1
    return dist



from __future__ import print_function
import inspect
import sys
import traceback


def pv(name):
    '''
    Snatched from:

    http://stackoverflow.com/questions/2813227/printing-variable-names-and-contents-as-debugging-tool-looking-for-emacs-python
    '''
    record = inspect.getouterframes(inspect.currentframe())[1]
    frame = record[0]
    val = eval(name, frame.f_globals, frame.f_locals)
    print('{0}: {1}'.format(name, val), file=sys.stderr)


def bt():
    '''
    Print the calls stack.
    '''
    for line in traceback.format_stack()[:-1]:
        print(line.strip())



#!/usr/bin/python
from __future__ import division
from builtins import map
from builtins import range
import timeit
import sys
import itertools
from collections import Counter

import forgi.utilities.debug as fud
import numpy as np
import math
import numpy.linalg as nl
import numpy.testing as nt
import random as rand
import warnings
import logging
#import scipy as sp

log = logging.getLogger(__name__)
try:
    profile
except:
    def profile(f):
        return f

# Set this module-level variable to False to disable all asserts in this module
USE_ASSERTS = __debug__

null_array = np.array([0., 0., 0.])

x_array = np.array([1., 0., 0.])
y_array = np.array([0., 1., 0.])
z_array = np.array([0., 0., 1.])

# identity matrix
identity_matrix = np.array([x_array, y_array, z_array])

standard_basis = np.array([[1., 0., 0.], [0., 1., 0.], [0., 0., 1.]])
tau = 2 * math.pi

# Seems to be unused


def get_inter_distances(vecs):
    '''
    Calculate all of the distances between the points of vecs.

    :param vecs: a list of vectors (=points)
    :return: a list containing all distances between any two vectors in vecs.
    '''
    distances = []
    for i in range(len(vecs)):
        for j in range(i + 1, len(vecs)):
            distances += [vec_distance(vecs[i], vecs[j])]

    return distances


def get_random_vector(mult=1.):
    """
    Returns a random vector.

    :param mult: Stretch the random vector by this value. This is the longest value allowed for the total length.
    :return: A random vector
    """
    # Using rejection sampling to generate uniform distribution from points inside a sphere.
    # Thanks to http://stats.stackexchange.com/a/7984/90399
    while True:
        vec = np.array([mult * rand.uniform(-1, 1), mult *
                        rand.uniform(-1, 1), mult * rand.uniform(-1, 1)])
        if magnitude(vec) <= mult:
            return vec


def get_orthogonal_unit_vector(vec):
    '''
    Return a vector orthogonal to vec.

    .. note::

        To create a basis, use create_orthonormal_basis instead!
    '''
    vec2 = get_non_colinear_unit_vector(vec)
    vec3 = np.cross(vec, vec2)
    return normalize(vec3)


# def get_random_vector_pair(angle=rand.uniform(0, math.pi)) -> Removed, because it was never used.


def get_double_alignment_matrix(vp1, vp2):
    '''
    Align two sets of two vectors onto each other.

    :param vp1: A pair of two vectors.
    :param vp2: Another pair of two vectors.
    '''
    angle1 = vec_angle(vp1[0], vp1[1])
    angle2 = vec_angle(vp2[0], vp2[1])

    if USE_ASSERTS:
        nt.assert_allclose(angle1, angle2, rtol=1e-7, atol=1e-7)

    # Align the first two segments
    mat1 = get_alignment_matrix(vp1[0], vp2[0])

    # See where the second segment of the second set ends up
    # after the first alignment
    new_vp2_1 = np.dot(mat1, vp2[1])

    comp1 = np.cross(vp1[1], vp1[0])
    #comp1 = np.cross(vp1[0], vp1[1])
    # should be along the plane of vp1[0] and vp1[1]
    comp2 = np.cross(vp1[0], comp1)

    basis1 = create_orthonormal_basis(normalize(vp1[0]), normalize(comp2))
    rej2 = change_basis(new_vp2_1, basis1, standard_basis)

    angle = math.atan2(rej2[2], rej2[1])

    mat2 = rotation_matrix(vp1[0], angle)

    # return np.dot(mat1, mat2)
    return np.dot(mat2, mat1)


def get_alignment_matrix(vec1, vec2):
    '''
    Return a rotation matrix that will align vec1 along vec2.

    :param vec1: The target vector.
    :param vec2: The vector to be aligned.
    '''

    comp = np.cross(vec1, vec2)
    angle = vec_angle(vec1, vec2)

    return rotation_matrix(comp, angle)


def get_non_colinear_unit_vector(vec):
    '''
    Get a unit vector that does not lie on the line defined by vec.

    This is done by creating a vector along the least represented axis in vec.

    :param vec: The vector under consideration.
    :return: A vector along an axis.
    '''
    absvec = [abs(v) for v in vec]
    m = min(absvec)
    ind = absvec.index(m)
    unit = [0., 0., 0.]
    unit[ind] = 1.

    return np.array(unit)


def is_almost_parallel(vec1, vec2):
    """
    Are vec1 and vec2 almost parallel?
    Nothing is parallel to the zero vector!

    :returns: 1 if vec1 and vec2 are parallel,
             -1 if they are antiparallel,
             0 if they are neither.
    """
    CUTOFF = 10**-7
    vec2_clean = []
    for c in vec2:
        if abs(c) < CUTOFF:
            vec2_clean.append(float("nan"))
        else:
            vec2_clean.append(c)
    factors = np.asarray(vec1) / np.asarray(vec2_clean)
    factors = [f for f in factors if not np.isnan(f)]
    log.debug("vec1 %s, vec2 %s, fac %s", vec1, vec2, factors)
    if not factors:
        return 0  # vec2~[0,0,0]
    elif all(np.sign(factors) == np.sign(factors[0])) and all(abs(f - factors[0]) < CUTOFF for f in factors):
        log.debug("Signs: %s", np.sign(factors))
        return np.sign(factors[0])  # returns 0, if vec1==[0,0,0]
    else:
        return 0  # Not (anti-)parallel


'''
# OLD:
def is_almost_parallel(vec1, vec2):
    """
    Returns true, if two vectors are almost parallel

    Note that every vector is parallel to the zero vector.
    """
    CUTOFF=10**-9
    for i in range(len(vec1)):
        if abs(vec2[i])>CUTOFF:
            factor=vec1[i]/vec2[i]
            break
    else:
        log.debug("vec2 is zero-vector")
        return True # vec2 is Zero-vector
    log.debug("Factor is %f", factor)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        log.debug("vec1/vec2 = %s", [abs(vec1[i]/vec2[i]) for i in range(len(vec1))])
    return all((abs(vec1[i])<CUTOFF and abs(vec2[i])<CUTOFF) or (abs(vec2[i])>CUTOFF and abs(vec1[i]/vec2[i]-factor)<CUTOFF) for i in range(len(vec1)))
'''


def line_segments_collinearity(segment1, segment2):
    """
    Quantifies, how collinear (according to some measure) two line segments are.

    :param segment1, segment2: Each a tuple of vectors (start, end)
    """
    dir1 = normalize(segment1[1] - segment1[0])
    dir2 = normalize(segment2[1] - segment2[0])

    # Get the average direction of the two vectors, if oriented correctly.
    s1 = dir1 + dir2
    s2 = dir1 - dir2
    if magnitude(s1) > magnitude(s2):
        sum_vec = s1
    else:
        sum_vec = s2
    # Further more, the line should pass through the center of the
    # 4 points defining the line segments.
    points = np.array(segment1 + segment2)
    center = points.mean(axis=0)
    '''
    # Now plot for verification
    import matplotlib.pyplot as plt
    from mpl_toolkits.mplot3d import Axes3D
    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')

    log.info("Fiorst 2 points: X %s Y %s Z %s ", points[:2,0], points[:2,1], points[:2,2])
    plt.plot(points[:2,0], points[:2,1], points[:2,2],"ro-")
    plt.plot(points[2:,0], points[2:,1], points[2:,2],"bo-")

    line = np.array([center-3*sum_vec, center, center+3*sum_vec])
    plt.plot(line[:,0],line[:,1],line[:,2], "g-o")
    ax.set_xlim(-5,5)
    ax.set_ylim(-5,5)
    ax.set_zlim(-5,5)

    plt.show()'''

    # Now calculate the distances of the 4 points from the fitted line
    d_point_line = point_line_distance(points, center, sum_vec)
    assert len(d_point_line) == 4
    # Return an R**2 value, like or linear regression (1=collinear, <1 worse)
    return 1 - sum(d_point_line**2) / ((points - center)**2).sum()

    # Different version via SVD
    centered_points = points - center
    uu, dd, vv = np.linalg.svd(centered_points)  # , full_matrices=False)
    # Now vv[0] is the direction of the target line.
    # Calculate the deviation of the points from the target line.
    distances_from_line = np.linalg.norm(
        np.cross(vv[0], centered_points), axis=1) / magnitude(vv[0])
    r2 = 1 - (np.sum(distances_from_line**2) / np.sum(centered_points**2))
    return r2


def create_orthonormal_basis(vec1, vec2=None, vec3=None):
    '''
    Create an orthonormal basis using the provided vectors.

    If more than one is provided, it must be orthogonal to
    the others.

    '''
    if vec1 is not None and vec2 is not None and vec3 is None and not USE_ASSERTS:
        try:
            from . import cytvec
        except ImportError as e:
            warnings.warn(
                "Extention modules (cython code) not installed, using slower python version")
        else:
            return cytvec.create_orthonormal_basis(vec1, vec2)

    if vec2 is None:
        vec2 = get_non_colinear_unit_vector(vec1)
        vec2 = np.cross(vec1, vec2)
    else:
        if USE_ASSERTS:
            if round(vec_angle(vec2, vec1), 9) != round(math.pi / 2, 9):
                assert False, ("vec2 {} is not normal to vec1 {}! Angle is {} rad ({} degrees)".format(
                    vec2, vec1, vec_angle(vec2, vec1), math.degrees(vec_angle(vec2, vec1))))

    mag_vec1 = magnitude(vec1)
    if mag_vec1 == 0:
        raise ZeroDivisionError("vec 1 {}  has magnitude 0.".format(vec1))
    vec1 = vec1 / mag_vec1

    mag_vec2 = magnitude(vec2)
    if mag_vec2 == 0:
        raise ZeroDivisionError(
            "vec 2 has magnitude 0. vecs are so far {} and {} ".format(vec1, vec2))
    vec2 = vec2 / mag_vec2

    if vec3 is None:
        vec3 = np.cross(vec1, vec2)

    mag_vec3 = magnitude(vec3)
    if mag_vec3 == 0:
        raise ZeroDivisionError("vec 3 has magnitude 0. vecs are {}, {} and {}".format(
            repr(vec1), repr(vec2), repr(vec3)))
    vec3 = vec3 / mag_vec3

    return np.array([vec1, vec2, vec3])


"""
# Code used for comparing the fastes method of creating an orthonormal basis:
def create_orthonormal_basis1(vec1, vec2=None, vec3=None):
    '''
    Create an orthonormal basis using the provided vectors.

    If more than one is provided, it must be orthogonal to
    the others.
    '''
    if vec2 == None:
        vec2 = get_non_colinear_unit_vector(vec1)
    #else:
    #    nt.assert_allclose(np.dot(vec2, vec1), 0., rtol=1e-7, atol=1e-7)


    vec1 = normalize(np.array(vec1))
    vec2 = normalize(np.array(vec2))

    if vec3 == None:
        vec3 = np.cross(vec1, vec2)

    vec3 = normalize(vec3)

    return np.array([vec1, vec2, vec3])

def time_cob1():
    vec1 = get_random_vector()
    vec2 = get_random_vector()

    basis = create_orthonormal_basis1(vec1, vec2)

def time_cob2():
    vec1 = get_random_vector()
    vec2 = get_random_vector()

    basis = create_orthonormal_basis(vec1, vec2)

def time_cob():
    t1 = timeit.Timer("time_cob1()", "from forgi.threedee.utilities.vector import time_cob1")
    t2 = timeit.Timer("time_cob2()", "from forgi.threedee.utilities.vector import time_cob2")

    print "1: ", t1.repeat(number=100000) #prints [3.045403003692627, 3.0388529300689697, 3.0359420776367188]
    print "2: ", t2.repeat(number=100000) #prints [2.7473390102386475, 2.74338698387146, 2.731964111328125]
"""


def spherical_cartesian_to_polar(vec):
    '''
    Return a parameterization of a vector of 3 coordinates:

    x = r sin u cos v
    y = r sin u sin v
    z = r cos u

    0 <= u <= pi
    -pi <= v <= pi

    Where u is the polar angle and v is the azimuth angle.

    :param vec: A vector of 3 cartesian coordinates.
    :param fast: Do not assert correctnes of result.
    :return: (r, u, v)
    '''
    r = magnitude(vec)
    u = math.acos(vec[2] / r)
    v = math.atan2(vec[1], vec[0])

    if USE_ASSERTS:
        nt.assert_allclose(vec[0], r * math.sin(u) *
                           math.cos(v), rtol=1e-7, atol=1e-7)
    return np.array((r, u, v))


def spherical_polar_to_cartesian(vec):
    '''
    Convert spherical polar coordinates to cartesian coordinates:

    See the definition of spherical_cartesian_to_polar.

    :param vec: A vector of the 3 polar coordinates (r, u, v)
    :return: (x, y, z)
    '''
    (r, u, v) = vec

    x = r * math.sin(u) * math.cos(v)
    y = r * math.sin(u) * math.sin(v)
    z = r * math.cos(u)

    return np.array([x, y, z])


def get_standard_basis(dim):
    '''
    Get a standard basis for the given dimension.

    For 2D, this equals [[1.,0.],[0.,1.]]

    :param dim: The dimension of the vector space.
    :return: A vector of vectors that constitute the standard basis.
    '''

    standard_basis = [[0. for j in range(dim)] for i in range(dim)]
    for i in range(dim):
        standard_basis[i][i] = 1.
    standard_basis = np.array(standard_basis)

    return standard_basis


def change_basis(coords, new_basis, old_basis):
    '''
    Change the basis of coordinates to a new basis. In a regular structure
    we have the coordinates in the regular cartesian coordinate system. For helix-helix
    orientations, however, we want to express the coordinates in a coordinate system
    defined by the first helix.

    The new basis will consist of the axis of the first helix, one of its twist elements
    (the one closest to the second vector) and a third vector orthogonal to the previous
    two.

    # http://tutorial.math.lamar.edu/Classes/LinAlg/ChangeOfBasis.aspx

    :param coords: The coordinates to transform (array of n elements).
    :param new_basis: The new basis vectors (n x n matrix)
    :param old_basis: The old basis for the coordinates(n x n matrix)
    :return: The new coordinates according to the new basis
    '''
    #assert(len(coords) == len(new_basis))
    #assert(len(new_basis) == len(old_basis))

    dim = len(coords)
    standard_coords = np.dot(old_basis.transpose(), coords)
    new_coords = nl.solve(new_basis.transpose(), standard_coords)

    return new_coords


def change_basis_vectorized(coords, new_basis, old_basis):
    """
    Change an array of vectors (coords) from old_basis to new_basis.

    :param coords: A array of coordinates to transform.
    """
    standard_coords = np.tensordot(
        old_basis.transpose(), coords, axes=([-1], [1])).T
    standard_to_new = nl.inv(new_basis.transpose())
    new_coords = np.tensordot(
        standard_to_new, standard_coords, axes=([-1], [1])).T
    return new_coords


""" # CODE USED FOR BENCHMARKING
    # change_basis1 is slightly faster or the same (4.61 vs 4.67)
def change_basis1(coords, new_basis, old_basis):
    '''
    '''

    dim = len(coords)
    standard_coords = np.dot(old_basis.transpose(), coords)
    standard_to_new = nl.inv(new_basis.transpose())
    new_coords = np.dot(standard_to_new, standard_coords)

    return new_coords


def change_basis2(coords, new_basis, old_basis):
    '''
    '''

    dim = len(coords)
    standard_coords = np.dot(old_basis.T, coords)
    new_coords = nl.solve(new_basis.T, standard_coords)

    return new_coords

def change_basis1_benchmark():
    coords = get_random_vector(10.)
    basis1 = np.array([get_random_vector(10.) for i in range(3)])
    basis2 = np.array([get_random_vector(10.) for i in range(3)])

    nc = change_basis1(coords, basis1, basis2)

def change_basis2_benchmark():
    coords = get_random_vector(10.)
    basis1 = np.array([get_random_vector(10.) for i in range(3)])
    basis2 = np.array([get_random_vector(10.) for i in range(3)])

    nc = change_basis2(coords, basis1, basis2)

def time_basis1():
    t1 = timeit.Timer("change_basis1_benchmark()","from forgi.threedee.utilities.vector import change_basis1_benchmark")
    print t1.repeat(number=100000)

def time_basis2():
    t2 = timeit.Timer("change_basis2_benchmark()","from forgi.threedee.utilities.vector import change_basis2_benchmark")
    print t2.repeat(number=100000)
"""


def vector_rejection(a, b):
    '''
    Return the vector rejection of a from b. In other words, return the orthogonal
    projection of a onto the plane orthogonal to b.

    :param a: The vector to be projected.
    :param b: The vector defining the normal of the plane.
    :return: The rejection of the vector a from b. (a - (np.dot(a, b) / np.dot(b, b)) * b)
    '''

    n = np.dot(a, b)
    d = np.dot(b, b)
    return a - (n / d) * b


def rotation_matrix(axis, theta):
    #TODO:  Use rotation_matrix_weave code (deleted in the commit that introduced this comment)
    # as a guide how to implement this in cython in the future for speedup.
    '''
    Calculate the rotation matrix for a CLOCKWISE rotation of theta around axis.
    This is in the opposite direction that is usually used.

    Thanks to unutbu on StackOverflow

    http://stackoverflow.com/questions/6802577/python-rotation-of-3d-vector

    :param axis: The axis around which to rotate. A np-array with length 3 or
                 one of "x", "y" and "z", where x:=standard_basis[0] etc.
    :param theta: The angle of rotation (in rad)
    :return: A matrix which can be used to perform the given rotation. The coordinates
             need only be multiplied by the matrix. (np.dot(matrix, vec))
    '''
    if isinstance(axis, (np.ndarray, list)):
        axis = normalize(axis)
        b, c, d = -axis * math.sin(theta / 2)
        a = math.cos(theta / 2)
        return np.array([[a * a + b * b - c * c - d * d, 2 * (b * c - a * d), 2 * (b * d + a * c)],
                         [2 * (b * c + a * d), a * a + c * c -
                          b * b - d * d, 2 * (c * d - a * b)],
                         [2 * (b * d - a * c), 2 * (c * d + a * b), a * a + d * d - b * b - c * c]])
    else:
        s = math.sin(theta)
        a = math.cos(theta)
        if axis == "y":
            return np.array([[a,  0, -s],
                             [0,  1,  0],
                             [s,  0,  a]])
        elif axis == "z":
            return np.array([[a,  s, 0],
                             [-s,  a, 0],
                             [0,  0, 1]])
        elif axis == "x":
            return np.array([[1,  0,  0],
                             [0,  a, s],
                             [0, -s,  a]])
        else:
            raise TypeError('Axis must be numpy array or one of "x", "y", "z"')


def _vector_set_rmsd(set1, set2):
    '''
    Calculate the not-centered rmsd between two sets of vectors.
    Currently in forgi.threedee.utilities.pdb, but subject to future deprecation.

    :param set1: A matrix
    :param set2: Another matrix.

    :return: The rmsd between the rows of the matrix.
    '''
    rmsd = 0
    count = 0
    for i in range(len(set1)):
        rmsd += magnitude(set2[i] - set1[i]) ** 2
        count += 1
    rmsd /= count
    return math.sqrt(rmsd)


def get_vector_centroid(crds1):
    '''
    Find the centroid of a set of vectors.

    :param crds: A matrix containing all of the vectors.

    :return: The centroid of the rows of the matrix crds.
    '''
    crds1 = np.asarray(crds1)
    centroid = np.sum(crds1, axis=0)
    centroid /= float(len(crds1))

    return centroid
    if crds1.shape[1] == 3:
        centroid1 = np.array([0., 0., 0.])
    else:
        centroid1 = np.array([0., 0.])

    for i in range(len(crds1)):
        centroid1 += crds1[i]

    centroid1 /= float(len(crds1))

    for i in centroid1:
        if math.isnan(i):
            raise ValueError('nan encountered in centroid: {}, len crds1 = {}.'.format(
                centroid1, len(crds1)))

    return centroid1


def center_on_centroid(crds1):
    centroid1 = get_vector_centroid(crds1)

    crds = np.asarray(crds1)
    return crds - centroid1


def magnitude(vec):
    '''
    Return the magnitude of a vector `(|V|)`.

    This is guaranteed for arbitrary dimensions of vec.

    :param vec: The vector in question.
    :return: The magnitude of the vector.
    '''
    # return np.linalg.norm(vec) #A lot of overhead, if used for a single vector
    return np.sqrt(np.dot(vec, vec))


def det3x3(matrix):
    """return the determinant of a 3x3 matrix"""
    positive = matrix[0, 0] * matrix[1, 1] * matrix[2, 2] + matrix[0, 1] * \
        matrix[1, 2] * matrix[2, 0] + \
        matrix[0, 2] * matrix[1, 0] * matrix[2, 1]
    minus = matrix[2, 0] * matrix[1, 1] * matrix[0, 2] + matrix[2, 1] * \
        matrix[1, 2] * matrix[0, 0] + \
        matrix[2, 2] * matrix[1, 0] * matrix[0, 1]
    return positive - minus


"""
def time_mag1():
    vec1 = get_random_vector()

    return math.sqrt(np.dot(vec1, vec1))

def time_mag2():
    vec1 = get_random_vector()

    return math.sqrt(np.dot(vec1, vec1))

def time_mag():
    t1 = timeit.Timer("time_mag1()", "from forgi.utilities.vector import time_mag1")
    t2 = timeit.Timer("time_mag2()", "from forgi.utilities.vector import time_mag2")

    print t1.repeat(number=10000)
    print t2.repeat(number=10000)
"""


def normalize(vec):
    '''
    Normalize a vector so that its magnitude becomes 1.0 while
    its direction remains the same.

    :param vec: The vector in question.
    :return: A normalized version of the vector.
    '''
    mag = magnitude(vec)
    if mag == 0:  # Numpy would return Nan and raise a RuntimeWarning.
        raise ValueError("Cannot normalize zero- vector!")
    return vec / mag


def vec_angle(vec1, vec2):
    '''
    Get the angle between two vectors using the identity:

    `A * B = |A||B| cos t`

    Where A and B are two vectors and t is the angle between themath.

    :param vec1: The first vector (A)
    :param vec2: The second vector (B)
    :return: The angle between the two vectors.
    '''

    vec1n = normalize(vec1)
    vec2n = normalize(vec2)

    d = np.dot(vec1n, vec2n)

    # this shouldn't happen, but sometimes it does, presumably because
    # of rounding errors
    if d >= 1.:
        d = 1.
    if d <= -1.:
        d = -1.

    angle = math.acos(d)
    return angle


def vec_dot(a, b):
    """
    Vector dot product for vectors of length 3.

    For small vectors of length 3 that are represented as lists and not as np.arary,
    this naive python implementation might be faster than the corresponding numpy implementation.

    If a and b are already numpy arrays, the numpy implementation seems to be faster
    (depending an how numpy was compiled)

    """
    # >>> timeit.timeit('ftuv.vec_dot(a,b)', setup='import numpy as np;import forgi.threedee.utilities.vector as ftuv; a=np.array([1,2,3]); b=np.array([7,2,5]);')
    # 0.7863450050354004
    # >>> timeit.timeit('ftuv.vec_dot(a,b)', setup='import numpy as np;import forgi.threedee.utilities.vector as ftuv; a=[1,2,3]; b=[7,2,5];')
    # 0.27366209030151367
    # >>> timeit.timeit('np.dot(a,b)', setup='import numpy as np; a=[1,2,3]; b=[7,2,5]')
    # 1.4038841724395752
    # >>> timeit.timeit('np.dot(a,b)', setup='import numpy as np; a=np.array([1,2,3]); b=np.array([7,2,5])')
    # 0.6194930076599121
    return a[0] * b[0] + a[1] * b[1] + a[2] * b[2]


'''
def cross(a, b):
    c = [a[1]*b[2] - a[2]*b[1],
         a[2]*b[0] - a[0]*b[2],
         a[0]*b[1] - a[1]*b[0]]

    return c
'''


def seg_intersect(line1, line2):
    """
    Intersection of 2 line segments in 2D space (as lists or numpy array-like).
    :param line1: a tuple/list (a1, a2): The first line segment, from a1 to a2
    :param line2: a tuple/list (b1, b2):The 2nd line segment, from b1 to b2
    """
    a1, a2 = line1
    b1, b2 = line2
    a1 = np.array(a1)
    a2 = np.array(a2)
    b1 = np.array(b1)
    b2 = np.array(b2)
    if max(map(len, [a1, a2, b1, b2])) != 2:
        raise ValueError(
            "Expecting only 2-dimensional vectors. Found higher-dimensional vector: a1={}, a2={}, b1={}, b2={}".format(a1, a2, b1, b2))
    if min(map(len, [a1, a2, b1, b2])) != 2:
        raise ValueError(
            "Expecting only 2-dimensional vectors. Found lower-dimensional vector.")
    if (a1 == a2).all() or (b1 == b2).all():
        raise ValueError(
            "Start and end of a line must not be equal! a1={}, a2={}, b1={}, b2={}".format(a1, a2, b1, b2))
    dxa = a2[0] - a1[0]
    dya = a2[1] - a1[1]
    dxb = b2[0] - b1[0]
    dyb = b2[1] - b1[1]
    num = a1[0] * dya - a1[1] * dxa - b1[0] * dya + b1[1] * dxa
    denom = float(dxb * dya - dyb * dxa)

    if denom == 0:
        # parallel or on same lines
        if dxa == 0:
            t1 = (b1[1] - a1[1]) / dya
            t2 = (b2[1] - a1[1]) / dya
            t1test = t1
        else:
            t1 = (b1[0] - a1[0]) / dxa
            t2 = (b2[0] - a1[0]) / dxa
        if dya == 0:
            t1test = t1
        else:
            t1test = (b1[1] - a1[1]) / dya
        if t1 != t1test:
            return []
        # On same line
        if dxa == 0:
            s1 = (a1[1] - b1[1]) / dyb
            s2 = (a2[1] - b1[1]) / dyb
        else:
            s1 = (a1[0] - b1[0]) / dxb
            s2 = (a2[0] - b1[0]) / dxb
        if all(x < 0 or x > 1 for x in [s1, s2, t1, t2]):
            return []
        ts = min(t1, t2)
        te = max(t1, t2)
        toret = []
        if ts < 0:
            toret.append(a1)
        else:
            toret.append(b1)
        if te < 1:
            toret.append(b2)
        else:
            toret.append(a2)
        return toret
    else:
        s = num / denom
        if s >= 0 and s <= 1:
            c = np.array(b1) + s * (np.array(b2) - np.array(b1))
            if dxa != 0:
                t = (c[0] - a1[0]) / dxa
            else:
                t = (c[1] - a1[1]) / dya
            if t >= 0 and t <= 1:
                return [c]
        return []


def point_line_distance(point, line_start, line_dir):
    """
    Calculate the distance between the point and the line
    through line_point with direction line_dir.

    :param point: A point(an array with shape (3,)) or multiple points (shape n,3)
    """
    if np.shape(point) == (3,):
        return magnitude(np.cross(line_dir, (line_start - point))) / magnitude(line_dir)
    else:
        # More than one point
        return np.linalg.norm(np.cross(line_dir, line_start - point), axis=1) / magnitude(line_dir)


def vec_distance(vec1, vec2):
    """
    The (euclidean) distance between two points vec1 and vec2.

    This is guaranteed to work for arbitrary but equal dimensions of vec1 and vec2.

    :param vec1, vec2: A list or np.array of floats
    :returns: A float
    """
    vec1 = np.asarray(vec1)
    vec2 = np.asarray(vec2)
    direction = vec2 - vec1
    return math.sqrt(np.dot(direction, direction))


@profile
def elements_closer_than(s1_p0, s1_p1, s2_p0, s2_p1, distance):
    '''
    Code copied from line_segment_distance, but with optimizations for fast comparison to distance.

    Code shamelessly translated from:
    http://softsurfer.com/Archive/algorithm_0106/algorithm_0106.htm#dist3D_Segment_to_Segment

    :param s1_p0: The start of the first segment
    :param s1_p1: The end of the first segment

    :param s2_p0: The start of the second segment
    :param s2_p1: The end of the second segment

    :return: True or False
    '''
    # pylint: disable=E1130  # See https://github.com/PyCQA/pylint/issues/2721
    u = s1_p1 - s1_p0
    v = s2_p1 - s2_p0
    w = s1_p0 - s2_p0
    lenw = magnitude(w)
    a = np.dot(u, u)        # always >= 0
    c = np.dot(v, v)        # always >= 0

    if lenw < distance:
        return True
    if lenw > math.sqrt(a) + math.sqrt(c) + distance:
        return False

    b = np.dot(u, v)

    d = np.dot(u, w)
    e = np.dot(v, w)

    D = a * c - b * b       # always >= 0
    sD = D      # sc = sN / sD, default sD = D >= 0
    tD = D      # tc = tN / tD, default tD = D >= 0

    SMALL_NUM = 0.000001

    # compute the line parameters of the two closest points
    if (D < SMALL_NUM):  # the lines are almost parallel
        sN = 0.0        # force using point P0 on segment S1
        sD = 1.0        # to prevent possible division by 0.0 later
        tN = e
        tD = c
    else:                # get the closest points on the infinite lines
        sN = (b * e - c * d)
        tN = (a * e - b * d)
        if (sN < 0.0):      # sc < 0 => the s=0 edge is visible
            sN = 0.0
            tN = e
            tD = c
        elif (sN > sD):  # sc > 1 => the s=1 edge is visible
            sN = sD
            tN = e + b
            tD = c

    if (tN < 0.0):           # tc < 0 => the t=0 edge is visible
        tN = 0.0
        # recompute sc for this edge
        if (-d < 0.0):
            sN = 0.0
        elif (-d > a):
            sN = sD
        else:
            sN = -d
            sD = a
    elif (tN > tD):      # tc > 1 => the t=1 edge is visible
        tN = tD
        # recompute sc for this edge
        if ((-d + b) < 0.0):
            sN = 0
        elif ((-d + b) > a):
            sN = sD
        else:
            sN = (-d + b)
            sD = a

    # finally do the division to get sc and tc
    sc = 0.0 if abs(sN) < SMALL_NUM else sN / sD
    tc = 0.0 if abs(tN) < SMALL_NUM else tN / tD

    # get the difference of the two closest points
    # dP = w + (sc * u) - (tc * v)  # = S1(sc) - S2(tc)

    return vec_distance(s1_p0 + sc * u, s2_p0 + tc * v) < distance


def line_segment_distance(s1_p0, s1_p1, s2_p0, s2_p1):
    '''
    Calculate the two points on each of the segments that are closest to
    each other. The first segment is defined as p1->p2 and the second as
    p3->p4.

    Code shamelessly translated from:
    http://softsurfer.com/Archive/algorithm_0106/algorithm_0106.htm#dist3D_Segment_to_Segment

    :param s1_p0: The start of the first segment
    :param s1_p1: The end of the first segment

    :param s2_p0: The start of the second segment
    :param s2_p1: The end of the second segment

    :return: A tuple of points (i1,i2) containing the point i1 on s1
        closest to the point i2 on segment s2.
    '''
    # pylint: disable=E1130  # See https://github.com/PyCQA/pylint/issues/2721
    u = s1_p1 - s1_p0
    v = s2_p1 - s2_p0
    w = s1_p0 - s2_p0

    a = np.dot(u, u)        # always >= 0
    b = np.dot(u, v)
    c = np.dot(v, v)        # always >= 0
    d = np.dot(u, w)
    e = np.dot(v, w)

    D = a * c - b * b       # always >= 0
    sD = D      # sc = sN / sD, default sD = D >= 0
    tD = D      # tc = tN / tD, default tD = D >= 0

    SMALL_NUM = 0.000001

    # compute the line parameters of the two closest points
    if (D < SMALL_NUM):  # the lines are almost parallel
        sN = 0.0        # force using point P0 on segment S1
        sD = 1.0        # to prevent possible division by 0.0 later
        tN = e
        tD = c
    else:                # get the closest points on the infinite lines
        sN = (b * e - c * d)
        tN = (a * e - b * d)
        if (sN < 0.0):      # sc < 0 => the s=0 edge is visible
            sN = 0.0
            tN = e
            tD = c
        elif (sN > sD):  # sc > 1 => the s=1 edge is visible
            sN = sD
            tN = e + b
            tD = c

    if (tN < 0.0):           # tc < 0 => the t=0 edge is visible
        tN = 0.0
        # recompute sc for this edge
        if (-d < 0.0):
            sN = 0.0
        elif (-d > a):
            sN = sD
        else:
            sN = -d
            sD = a
    elif (tN > tD):      # tc > 1 => the t=1 edge is visible
        tN = tD
        # recompute sc for this edge
        if ((-d + b) < 0.0):
            sN = 0
        elif ((-d + b) > a):
            sN = sD
        else:
            sN = (-d + b)
            sD = a

    # finally do the division to get sc and tc
    sc = 0.0 if abs(sN) < SMALL_NUM else sN / sD
    tc = 0.0 if abs(tN) < SMALL_NUM else tN / tD

    # get the difference of the two closest points
    # dP = w + (sc * u) - (tc * v)  # = S1(sc) - S2(tc)

    return (s1_p0 + sc * u, s2_p0 + tc * v)


def closest_point_on_seg(seg_a, seg_b, circ_pos):
    '''
    Closest point between a line segment and a point.

    Lifted from:

    http://doswa.com/2009/07/13/circle-segment-intersectioncollision.html
    '''
    if not isinstance(seg_a, np.ndarray):
        seg_a = np.array(seg_a)
        seg_b = np.array(seg_b)
        circ_pos = np.array(circ_pos)
    seg_v = seg_b - seg_a
    pt_v = circ_pos - seg_a
    mag = math.sqrt(sum(seg_v * seg_v))

    if mag <= 0:
        raise ValueError("Invalid segment length")
    seg_v_unit = seg_v / mag
    proj = pt_v.dot(seg_v_unit)
    if proj <= 0:
        return seg_a.copy()
    if proj >= mag:
        return seg_b.copy()
    proj_v = seg_v_unit * proj
    closest = proj_v + seg_a
    return closest


def cylinder_line_intersection(cyl, line, r):
    '''
    Get the points of intersection between a line and a cylinder.

    If they do not intersect, return an empty list. If the line
    touches the cylinder, then return a 2 point list with two identical points.
    If the line crosses the cylinder, then return a list of 2 points.
    '''

    cyl = np.array(cyl)
    line = np.array(line)
    cyl_vec = cyl[1] - cyl[0]
    #line_vec = line[1] - line[0]

    cyl_basis = create_orthonormal_basis(cyl_vec)

    line_t = change_basis(line.T, cyl_basis, standard_basis).T
    cyl_t = change_basis(cyl.T, cyl_basis, standard_basis).T

    #cyl_vec_t = cyl_t[1] - cyl_t[0]
    line_vec_t = line_t[1] - line_t[0]

    line_vec_t_normed = normalize(line_vec_t)

    # get the point on the line closest to the
    # center of the cylinder
    p = closest_point_on_seg(line_t[0][1:], line_t[1][1:],
                             np.array(cyl_t[0][1:]))

    if line_vec_t[1] == 0:
        return []

    # Figure out the x position by determining how far along
    # the y-coordinate of the segment the closest point is
    x = (line_t[0][0] +
         (line_vec_t[0] *
          (p[0] - line_t[0][1]) / line_vec_t[1]))
    v = p - cyl_t[0][1:]
    o = math.sqrt(sum(v * v))
    p = [x, p[0], p[1]]

    if o > r:
        # no intersection
        return []

    t = math.sqrt(r ** 2 - o ** 2)

    i1 = p - t * line_vec_t_normed
    i2 = p + t * line_vec_t_normed

    endpoints_t = [i1, i2]

    endpoints_t = sorted(endpoints_t, key=lambda x: x[0])
    line_t = sorted(line_t, key=lambda x: x[0])
    cyl_t = sorted(cyl_t, key=lambda x: x[0])

    # the start point will be the highest of the top of the cylinder
    # the end of the line, and the point where the line intersects
    # the surface of the cylinder
    start_points = np.array([cyl_t[0][0], line_t[0][0], endpoints_t[0][0]])

    # the end point will be the lowest of the... " " "
    end_points = np.array([cyl_t[1][0], line_t[1][0], endpoints_t[1][0]])

    start_points = sorted(start_points)
    end_points = sorted(end_points)

    start_param = (start_points[-1] - p[0]) / line_vec_t_normed[0]
    end_param = (end_points[0] - p[0]) / line_vec_t_normed[0]

    endpoints_t[0] = p + start_param * line_vec_t_normed
    endpoints_t[1] = p + end_param * line_vec_t_normed

    '''
    endpoints_t[0] = [start_points[-1],
                      endpoints_t[
    endpoints_t[1][0] = end_points[0]
    real_start = start_points[-1]
    real_end = end_points[0]
    '''

    intersects_t = np.array(endpoints_t)

    if intersects_t[0][0] > intersects_t[1][0]:
        # degenerate case, the line is to the side of the cylinder
        return []

    return change_basis(intersects_t.T, standard_basis, cyl_basis).T

# COVERAGE: Not used in ernwin and forgi at least since forgi v0.3


def pin_fits_two_cyl(cyl1, cyl2, cyl_width):
    '''
    If we create a cone that starts at one end of cylinder1, does it
    enclose cylinder2?

    This function approximates an answer by projecting a circle on
    the plane normal to the axis of cylinder1.

    :param cyl1: The coordinates of cylinder1
    :param cyl2: The coordinates of cylinder2
    :param cyl_width: The widths of the two cylinders
    :return: True or False
    '''
    basis = create_orthonormal_basis(cyl1[1] - cyl1[0])
    cyl2 = np.array([cyl2[0] - cyl1[1], cyl2[1] - cyl1[1]])

    cyl2_t = change_basis(cyl2.T, basis, standard_basis).T
    cone_width = cyl_width
    cyl1_len = magnitude(cyl1[1] - cyl1[0])

    # the cone expands
    cone_width_cyl2_start = cyl_width * (cyl1_len + cyl2_t[0][0]) / cyl1_len
    cone_width_cyl2_end = cyl_width * (cyl1_len + cyl2_t[1][0]) / cyl1_len

    cyl2_start_offset = math.sqrt(cyl2_t[0][1] ** 2 + cyl2_t[0][2] ** 2)
    cyl2_end_offset = math.sqrt(cyl2_t[1][1] ** 2 + cyl2_t[1][2] ** 2)

    if cyl2_start_offset > cone_width_cyl2_start:
        return False
    if cyl2_end_offset > cone_width_cyl2_end:
        return False

    return True

# COVERAGE: Not used in ernwin and forgi at least since forgi v0.3


def point_in_cylinder(pt1, pt2, r, testpt):
    '''
    Determine if testpt is within a cylinder of radius r.

    Translated from:

        http://www.flipcode.com/archives/Fast_Point-In-Cylinder_Test.shtml

    :param pt1: The start of the cylinder axis.
    :param pt2: The end of the cylinder axis.
    :param r: The radius of the cylinder.
    :param testpt: The point we are testing.
    :return: True if the point is within the cylinder, False otherwise.
    '''
    d = pt2 - pt1
    lengthsq = np.dot(d, d)
    radius_sq = r * r

    pd = testpt - pt1
    dot = np.dot(pd, d)

    if dot < 0. or dot > lengthsq:
        # beyond the edges of the cylinder
        return False
    else:
        dsq = np.dot(pd, pd) - (dot * dot) / lengthsq

        if (dsq > radius_sq):
            return False
        else:
            return True


def GetPointsEquiAngularlyDistancedOnSphere(numberOfPoints=45):
    """ each point you get will be of form 'x, y, z'; in cartesian coordinates
        eg. the 'l2 distance' from the origion [0., 0., 0.] for each point will be 1.0
        ------------
        converted from:  http://web.archive.org/web/20120421191837/http://www.cgafaq.info/wiki/Evenly_distributed_points_on_sphere )
    """
    dlong = math.pi * (3.0 - math.sqrt(5.0))  # ~2.39996323
    dz = 2.0 / numberOfPoints
    long = 0.0
    z = 1.0 - dz / 2.0
    ptsOnSphere = []
    for k in range(0, numberOfPoints):
        r = math.sqrt(1.0 - z * z)
        ptNew = (math.cos(long) * r, math.sin(long) * r, z)
        ptsOnSphere.append(ptNew)
        z = z - dz
        long = long + dlong

    return ptsOnSphere


def sortAlongLine(start, end, points):
    """
    Sort all points in points along the line from start to end.

    :param start: A point
    :param end: A point
    :param points: A list of points

    :returns: A list containing start, end and all elements of points, sorted by the distance from start
    """
    # print start, end, points
    s_points = points + [start, end]
    s_points.sort(key=lambda x: vec_distance(x, start))
    return s_points


def middlepoint(vec1, vec2):
    """The point in the middle between vec1 and vec2."""
    generator = ((x + vec2[i]) / 2.0 for i, x in enumerate(vec1))
    typ = type(vec1)
    if typ == np.ndarray:
        return np.fromiter(generator, float, len(vec1))
    return typ(generator)


def pair_distance_distribution(points, stepsize=1):
    return pair_distance_distribution_vectorized(points, stepsize)
    # The below is left in here as a not vectorized (slow) reference implementation
    dists = Counter()
    for p1, p2 in itertools.combinations(points, r=2):
        d = vec_distance(p1, p2)
        dists[d // stepsize] += 1
    m = max(dists.keys())
    x = np.arange(0, m + 1)
    return x * stepsize, np.array([dists[xi] for xi in x])

def pair_distance_distribution_vectorized(points, stepsize=1):
    dists = Counter()
    p1s = []
    p2s = []
    for p1, p2 in itertools.combinations(points, r=2):
        p1s.append(p1)
        p2s.append(p2)
    diffs=np.array(p1s)-np.array(p2s)
    lengths=np.sqrt(np.sum(diffs*diffs, axis=1))
    lengths = lengths // stepsize
    for d in lengths:
        dists[d] += 1
    m = max(dists.keys())
    x = np.arange(0, m + 1)
    return x * stepsize, np.array([dists[xi] for xi in x])



from __future__ import print_function, unicode_literals

from builtins import zip
from builtins import range
import sys
import re
import warnings
import itertools
import string
import math
import numpy as np
import Bio.PDB as bpdb
from collections import defaultdict

import forgi.utilities.debug as fud
import forgi.threedee.utilities.vector as ftuv
from forgi.threedee.utilities.modified_res import to_4_letter_alphabeth
import forgi.graph.residue as fgr

from logging_exceptions import log_to_exception

import logging
log = logging.getLogger(__name__)


class AtomName(str):
    """
    Like a string, but "C1'" and "C1*" compare equal
    """

    def __eq__(self, other):
        if self.endswith("*"):
            self = self[:-1] + "'"
        if other.endswith("*"):
            other = other[:-1] + "'"
        return str(self) == str(other)

    def __hash__(self):
        if self.endswith("*"):
            self = self[:-1] + "'"
        return hash(str(self))


backbone_atoms = list(map(AtomName, ['P', "O5'", "C5'", "C4'", "C3'", "O3'"]))
ring_atoms = list(map(AtomName, ["C4'", "C3'", "C2'", "C1'", "O4'", "O2'"]))

nonsidechain_atoms = backbone_atoms + ring_atoms

chi_torsion_atoms = dict()
chi_torsion_atoms['A'] = chi_torsion_atoms['G'] = list(
    map(AtomName, ["O4'", "C1'", "N9", "C4"]))
chi_torsion_atoms['C'] = chi_torsion_atoms['U'] = list(
    map(AtomName, ["O4'", "C1'", "N1", "C2"]))

side_chain_atoms = dict()
side_chain_atoms['U'] = list(
    map(AtomName, ['N1', 'C2', 'O2', 'N3', 'C4', 'O4', 'C5', 'C6']))
side_chain_atoms['C'] = list(
    map(AtomName, ['N1', 'C2', 'O2', 'N3', 'C4', 'N4', 'C5', 'C6']))

side_chain_atoms['A'] = list(
    map(AtomName, ['N1', 'C2', 'N3', 'C4', 'C5', 'C6', 'N6', 'N7', 'C8', 'N9']))
side_chain_atoms['G'] = list(
    map(AtomName, ['N1', 'C2', 'N2', 'N3', 'C4', 'C5', 'C6', 'O6', 'N7', 'C8', 'N9']))

all_side_chains = set(
    side_chain_atoms['U'] + side_chain_atoms['C'] + side_chain_atoms['A'] + side_chain_atoms['G'])

all_rna_atoms = set(nonsidechain_atoms) | all_side_chains

RNA_RESIDUES = ["A", "U", "G", "C", 'rA', 'rC', 'rG', 'rU', 'DU']

interactions = [(AtomName(a), AtomName(b)) for a, b in map(sorted,
                                                           [('P', "O5'"),
                                                            ('P', 'OP1'),
                                                            ('P', 'O1P'),
                                                            ('P', 'OP2'),
                                                            ('P', 'O2P'),
                                                            ("C2'", "O2'"),
                                                            ("O5'", "C5'"),
                                                            ("C5'", "C4'"),
                                                            ("C4'", "O4'"),
                                                            ("C4'", "C3'"),
                                                            ("O4'", "C1'"),
                                                            ("C3'", "C2'"),
                                                            ("C3'", "O3'"),
                                                            ("C2'", "C1'"),
                                                            ("C1'", "N1"),
                                                            ('N1', 'C2'),
                                                            ('N1', 'C6'),
                                                            ('C6', 'C5'),
                                                            ('C5', 'C4'),
                                                            ('C4', 'O4'),
                                                            ('C4', 'N4'),
                                                            ('C4', 'N3'),
                                                            ('N3', 'C2'),
                                                            ('C2', 'O2'),
                                                            ('C2', 'N2'),
                                                            ("C1'", "N9"),
                                                            ('N9', 'C8'),
                                                            ('N9', 'C4'),
                                                            ('C8', 'N7'),
                                                            ('N7', 'C5'),
                                                            ('C6', 'O6'),
                                                            ('C6', 'N6')])]



def rename_chains_for_pdb(chains):
    """
    :param chains: A dict chain_id:chain
    """
    used_chainids = set(chains.keys())
    def get_available_chainid():
        for c in string.ascii_uppercase:
            if c not in used_chainids:
                used_chainids.add(c)
                return c
        raise ValueError("Too many chains. Cannot convert to old PDB format.")
    for chain in chains.values():
        if len(chain.id)>1:
            chain.id = get_available_chainid()
    return {c.id: c for c in chains.values()}

def trim_chain_between(chain, start_res, end_res):
    '''
    Remove all nucleotides between start_res and end_res, inclusive.

    The chain is modified in place so there is no return value.
    '''
    to_detach = []
    for res in chain:
        if start_res <= res.id[1] and res.id[1] <= end_res:
            to_detach += [res]

    for res in to_detach:
        chain.detach_child(res.id)


def extract_subchains_from_seq_ids(all_chains, seq_ids):
    '''
    Extract a portion of one or more pdb chains.
    Creates a list of new chains which contain only
    the specified residues copied from the original chain.

    The chain ids are not modified.

    :param all_chains: A dictionary {chainid:chains}.
    :param seq_ids: An iterable of complete RESIDS.

    :returns: A dictionary chain-id:Bio.PDB.Chain.Chain objects
    '''
    new_chains = {}
    assert isinstance(all_chains, dict)
    for r in seq_ids:
        if r.chain in new_chains:
            chain = new_chains[r.chain]
        else:
            chain = new_chains[r.chain] = bpdb.Chain.Chain(r.chain)
        try:
            chain.add(all_chains[r.chain][r.resid].copy())
        except KeyError:
            log.info(list(sorted(all_chains[r.chain].child_dict.keys())))
            raise
    return new_chains


def is_covalent(contact):
    '''
    Determine if a particular contact is covalent.

    This does not look at the geometric distance but only at the atom names.

    :param contact: A pair of two Atom objects
    :return: `True` if they are covalently bonded
             `False` otherwise
    '''
    r1 = contact[0].parent
    r2 = contact[1].parent

    r1a = (r1, contact[0])
    r2a = (r2, contact[1])

    if contact[0].name.find('H') >= 0 or contact[1].name.find('H') >= 0:
        return True

    ((r1, c1), (r2, c2)) = sorted((r1a, r2a), key=lambda x: x[0].id[1])

    if r1.id == r2.id:
        if tuple(sorted((c1.name, c2.name))) in interactions:
            return True

    if r2.id[1] - r1.id[1] == 1:
        # neighboring residues
        if c1.name == 'O3*' and c2.name == 'P':
            return True

    return False


def num_noncovalent_clashes(chain):
    '''
    Check if a chain has non-covalent clashes. Non-covalent clashes are found
    when two atoms that aren't covalently linked are within 1.8 A of each other.

    :param chain: The chain to evaluate
    :param return: The number of non-covalent clashes.
    '''
    all_atoms = bpdb.Selection.unfold_entities(chain, 'A')
    ns = bpdb.NeighborSearch(all_atoms)

    contacts = ns.search_all(1.9)

    return len([c for c in contacts if not is_covalent(c)])


def noncovalent_distances(chain, cutoff=0.3):
    '''
    Print out the distances between all non-covalently bonded atoms
    which are closer than cutoff to each other.

    :param chain: The Bio.PDB chain.
    :param cutoff: The maximum distance
    '''
    all_atoms = bpdb.Selection.unfold_entities(chain, 'A')
    ns = bpdb.NeighborSearch(all_atoms)

    contacts = ns.search_all(cutoff)

    return [ftuv.magnitude(c[1] - c[0]) for c in contacts if not is_covalent(c)]



def pdb_rmsd(c1, c2, sidechains=False, superimpose=True ):
    '''
    Calculate the all-atom rmsd between two RNA chains.

    :param c1: A Bio.PDB.Chain
    :param c2: Another Bio.PDB.Chain
    :return: The rmsd between the locations of all the atoms in the chains.
    '''
    c1_list = [cr for cr in c1.get_list() if cr.resname.strip()
               in RNA_RESIDUES]
    c2_list = [cr for cr in c2.get_list() if cr.resname.strip()
               in RNA_RESIDUES]
    return residuelist_rmsd(c1_list, c2_list, sidechains, superimpose )

def residuelist_rmsd(c1_list, c2_list, sidechains=False, superimpose=True):
    import forgi.threedee.model.similarity as ftms

    if len(c1_list) != len(c2_list):
        log.error("c1_list (len %s): %s", c1_list)
        log.error("c2_list (len %s): %s", c2_list)
        raise Exception(
            "Chains of different length. (Maybe an RNA-DNA hybrid?)")

    #c1_list.sort(key=lambda x: x.id[1])
    #c2_list.sort(key=lambda x: x.id[1])
    to_residues = []
    crds1 = []
    crds2 = []
    all_atoms1 = []
    all_atoms2 = []
    for r1, r2 in zip(c1_list, c2_list):
        if sidechains:
            anames = nonsidechain_atoms + \
                side_chain_atoms[r1.resname.strip()]
        else:
            anames = nonsidechain_atoms
        #anames = a_5_names + a_3_names

        for a in anames:
            try:
                at1 = r1[a]
                at2 = r2[a]
            except:
                continue
            else:
                all_atoms1.append(at1)
                all_atoms2.append(at2)
                crds1.append(at1.coord)
                crds2.append(at2.coord)
                to_residues.append(r1)

    diff_vecs = ftms._pointwise_deviation(crds1, crds2)
    dev_per_res = defaultdict(list)
    for i, res in enumerate(to_residues):
        dev_per_res[res].append(diff_vecs[i])

    if superimpose:
        sup = bpdb.Superimposer()
        sup.set_atoms(all_atoms1, all_atoms2)

        return (len(all_atoms1), sup.rms, sup.rotran, dev_per_res)
    else:
        return (len(all_atoms1), ftuv._vector_set_rmsd(crds1, crds2), None, dev_per_res)


def get_first_chain(filename):
    '''
    Load a PDB file using the Bio.PDB module and return the first chain.

    :param filename: The path to the pdb file
    '''
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        s = bpdb.PDBParser(PERMISSIVE=False).get_structure('t', filename)
        return list(s.get_chains())[0]


def pdb_file_rmsd(fn1, fn2):
    '''
    Calculate the RMSD of all the atoms in two pdb structures.

    :param fn1: The first filename.
    :param fn2: The second filename.
    :return: The rmsd between the two structures.
    '''
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")

        s1 = bpdb.PDBParser().get_structure('t', fn1)
        s2 = bpdb.PDBParser().get_structure('t', fn2)

    c1, _, _ = get_biggest_chain(fn1)
    c2, _, _ = get_biggest_chain(fn2)

    rmsd = pdb_rmsd(c1, c2)

    return rmsd


def renumber_chain(chain, resids=None):
    '''
    Renumber all the residues in this chain so that they start at 1 and end at
    len(chain)

    :param chain: A Bio.PDB.Chain object
    :return: The same chain, but with renamed nucleotides
    '''

    counter = 1

    if resids is None:
        resids = [(' ', i + 1, ' ') for i in range(len(chain))]

    new_child_dict = dict()
    new_child_list = []

    for res, r_new in zip(chain, resids):
        res.id = r_new
        new_child_dict[res.id] = res
        new_child_list.append(res)

    chain.child_dict = new_child_dict
    chain.child_list = new_child_list

    return chain


def output_chain(chain, filename, fr=None, to=None):
    '''
    Dump a chain to an output file. Remove the hydrogen atoms.

    :param chain: The Bio.PDB.Chain to dump.
    :param filename: The place to dump it.
    '''
    class HSelect(bpdb.Select):
        def accept_atom(self, atom):
            if atom.name.find('H') >= 0:
                return False
            else:
                return True
    m = bpdb.Model.Model(' ')
    s = bpdb.Structure.Structure(' ')

    m.add(chain)
    s.add(m)

    io = bpdb.PDBIO()
    io.set_structure(s)
    io.save(filename, HSelect())


def output_multiple_chains(chains, filename, file_type="pdb"):
    '''
    Dump multiple chains to an output file. Remove the hydrogen atoms.

    :param chains: An iterable of Bio.PDB.Chain to dump.
    :param filename: The place to dump it.
    '''
    class HSelect(bpdb.Select):
        def accept_atom(self, atom):
            if atom.name.find('H') >= 0:
                return False
            else:
                return True
    m = bpdb.Model.Model(0)
    s = bpdb.Structure.Structure('stru')
    for chain in chains:
        log.debug("Adding chain %s with %s residues", chain.id, len(chain))
        m.add(chain)
        if file_type=="pdb" and len(chain.id)!=1:
            raise ValueError("Cannot save chain with name %s (not a single character) "
                             "in PDB format. Use cif format instead!")
    s.add(m)
    if file_type == "pdb":
        io = bpdb.PDBIO()
    else:
        io = bpdb.MMCIFIO()
    io.set_structure(s)
    try:
        io.save(filename, HSelect())
    except Exception as e:
        with log_to_exception(log, e):
            log.error("Could not output PDB with chains and residues:")
            for chain in s[0]:
                log.error("%s: %s", chain.id, [r.id for r in chain])
        raise


def get_particular_chain(in_filename, chain_id, parser=None):
    '''
    Load a PDB file and return a particular chain.

    :param in_filename: The name of the pdb file.
    :param chain_id: The id of the chain.
    :return: A Bio.PDB.Chain object containing that particular chain.
    '''
    chains, mr, ir = get_all_chains(in_filename, parser)
    chain, = [c for c in chains if c.id == chain_id]
    return chain, mr, ir


def get_biggest_chain(in_filename, parser=None):
    '''
    Load the PDB file located at filename, select the longest
    chain and return it.

    :param in_filename: The location of the original file.
    :return: A Bio.PDB chain structure corresponding to the longest
             chain in the structure stored in in_filename
    '''
    chains, mr, ir = get_all_chains(in_filename, parser)
    biggest = 0
    biggest_len = 0

    for i in range(len(chains)):
        c = chains[i]

        # Only count RNA residues
        num_residues = 0
        for res in c:
            if (res.resname.strip() == 'A' or
                res.resname.strip() == 'C' or
                res.resname.strip() == 'G' or
                    res.resname.strip() == 'U'):
                num_residues += 1

        if num_residues > biggest_len:
            biggest = i
            biggest_len = num_residues

    # sys.exit(1)

    orig_chain = chains[biggest]
    return orig_chain, mr, ir

def _extract_symmetrymatrices_from_cif_dict(cif_dict):
    """
    Code originally by G. Entzian, modified

    Extract matrices for symmetry operations from the cif-dict to
    convert the assymetric unit to the biological unit

    :returns: A tuple matrices, vectors.
              matrices is a dictionary {operation_id: rot_matrix}
              vectors a dictionary {operation_id: vector}
    """
    symmetry_ids = cif_dict["_pdbx_struct_oper_list.id"]
    matrices = defaultdict(lambda: [[0.0 for x in range(3)] for y in range(3)])
    vectors = defaultdict(lambda: [0.0 for x in range(3)])
    for i in range(1,4):
        v_key = '_pdbx_struct_oper_list.vector['+str(i)+']'
        value_vectors = cif_dict[v_key]
        if not isinstance(value_vectors, list):
            value_vectors = [value_vectors]
        for k, v in enumerate(value_vectors):
            vectors[symmetry_ids[k]][i-1] = float(v)

        for j in range(1,4):
            key = '_pdbx_struct_oper_list.matrix['+str(i)+']['+str(j)+']'
            value_per_matrix = cif_dict[key]
            if not isinstance(value_per_matrix, list):
                value_per_matrix = [value_per_matrix]
            for k, v in enumerate(value_per_matrix):
                matrices[symmetry_ids[k]][i-1][j-1] = float(v)
    return matrices, vectors

def _convert_cif_operation_id_expression(expression):
    """
    By Gregor Entzian
    """
    tmp_expr = expression
    op_id_list = []
    match = re.findall("\([\w\-\,\d]+\)", tmp_expr)#("(\d+)\-(\d+)", id)
    if match:
        for m in match:
            match_s = re.search("\((\d+)\-(\d+)\)", m)
            if match_s:
                f = match_s.group(1)
                t = match_s.group(2)
                f = int(f)
                t = int(t)
                for n in range(f,t+1):
                    op_id_list.append(str(n))
                tmp_expr = tmp_expr.replace(m,",")
                continue
            else:
                #split ',' in case of more entries. In case of one entry use the same code.
                l = m[1:-1] # remove parentheses
                parts = l.split(',')
                for p in parts:
                    op_id_list.append(str(p))
                tmp_expr = tmp_expr.replace(m,",")
                continue
    if tmp_expr != None and tmp_expr != "":
        tmp_expr = tmp_expr.replace("(","").replace(")","")
        parts = tmp_expr.split(',')
        for p in parts:
            if ',' not in p and p != '':
                match_s = re.search("(\d+)\-(\d+)", p)
                if match_s:
                    f = match_s.group(1)
                    t = match_s.group(2)
                    f = int(f)
                    t = int(t)
                    for n in range(f,t+1):
                        op_id_list.append(str(n))
                    continue
                op_id_list.append(str(p))
    return tuple(op_id_list)

def _extract_assembly_gen(cif_dict):
    """
    Extracts the information, how assemblies are generated out
    of the chains and symmetry operations, from the cifdict.

    :returns: A dictionary {assembly_id: [(chainid, operationids)...]}
              Where operation_ids is a list
    """
    assembly_ids = cif_dict['_pdbx_struct_assembly_gen.assembly_id']       #1
    operation_ids = cif_dict['_pdbx_struct_assembly_gen.oper_expression']   #1,2
    chain_id_lists = cif_dict['_pdbx_struct_assembly_gen.asym_id_list']      #A,B

    if not isinstance(assembly_ids, list):
        assembly_ids=[assembly_ids]
        assert not isinstance(operation_ids, list)
        operation_ids=[operation_ids]
        assert not isinstance(chain_id_lists, list)
        chain_id_lists=[chain_id_lists]

    assembly_components = defaultdict(lambda:defaultdict(list)) # assembly: operation-lists: chains
    for i, aid in enumerate(assembly_ids):
        op_ids=_convert_cif_operation_id_expression(operation_ids[i])
        chain_ids=chain_id_lists[i].split()
        assembly_components[0][operation_ids].extend(chain_id_lists)
    return assembly_components

def _get_new_chainid(chain_id, op_id, taken_ids):
    new_id= chain_id+"op"+str(op_id)
    while new_id in taken_ids:
        new_id+="a"
    return new_id

def _extract_asym_auth_id_map(cif_dict):
    label_ids = cif_dict['_atom_site.label_asym_id']
    auth_ids = cif_dict['_atom_site.auth_asym_id']
    l2a=defaultdict(lambda: None)
    a2l=defaultdict(list)
    for i, lid in enumerate(label_ids):
        if lid in l2a:
            if l2a[lid]!=auth_ids[i]:
                log.error("lid %s, authid %s but before %s", lid, auth_ids[i], l2a[lid])
                raise ValueError("Inconsistent cif.")
        else:
            l2a[lid]=auth_ids[i]
            a2l[auth_ids[i]].append(lid)
    log.debug(l2a)
    return l2a, a2l

def _get_assemblies(chains, cifdict):
    assemblies = _extract_assembly_gen(cifdict)
    id2chainid, chainid2ids = _extract_asym_auth_id_map(cifdict)
    if assembly_nr is None:
        assembly_nr = list(sorted(assemblies.keys()))[0]
    assembly_gen = assemblies[assembly_nr]
    old_chains = { c.id:c for c in chains}
    new_chains = {}
    log.debug("Original chains are %s. Now performing symmetry "
              "for assembly %s", old_chains, assembly_nr)
    log.debug("AG: %s",assembly_gen)
    for operations, labelids in assembly_gen[0].items():
        chainids = set(id2chainid[lid] for lid in labelids)
        lids_back = set( x for x in a2l for cid in chainids for a2l in chainid2ids[cid])
        if lids_back!=set(labelids):
            log.error("%s, %s, extra: %s", lids_back, labelids, lids_back-set(labelids))
            raise ValueError("Operation on part of an author designated assymetric unit "
                             "(auth_asym_id) nt supported.")
        for chain_id in chainids:
            if chain_id not in old_chains:
                log.debug ("Skipping chain %s: not RNA? chains: %s", chain_id, old_chains.keys())
                continue
            for op_id in op_ids:
                log.debug("Applying op %s to %s", op_id, chain_id)
                if chain_id in new_chains:
                    chain = old_chains[chain_id].copy()
                    newid = _get_new_chainid(chain_id, op_id,
                                             set(old_chains.keys())| set(new_chains.keys()))
                    log.debug("Setting id %s", newid)
                    chain.id = newid
                    chain.transform(operation_mat[op_id], operation_vec[op_id])
                new_chains[chain.id]=chain
    log.debug("new_chains: %s", new_chains)
    chains = list(new_chains.values())
    return chains

def get_all_chains(in_filename, parser=None, no_annotation=False, assembly_nr=None):
    '''
    Load the PDB file located at filename, read all chains and return them.

    :param in_filename: The location of the original file.
    :param assembly_nr: Which assembly to return. Default: The first.
    :return: a tuple chains, missing_residues

             * chains: A list of Bio.PDB chain structures corresponding to all
                       RNA structures stored in in_filename
             * missing_residues: A list of dictionaries, describing the missing residues.
             * interacting residues: A list of residues
    '''
    if parser is None:
        if in_filename.endswith(".pdb"):
            parser = bpdb.PDBParser()
        elif in_filename.endswith(".cif"):
            parser = bpdb.MMCIFParser()
        else:  # Cannot determine filetype by extention. Try to read first line.
            with open(in_filename) as pdbfile:
                line = pdbfile.readline(20)
                # According to
                # page 10 of ftp://ftp.wwpdb.org/pub/pdb/doc/format_descriptions/Format_v33_A4.pdf
                # a HEADER entry is mandatory. Biopython sometime starts directly with ATOM
                if line.startswith("HEADER") or line.startswith("ATOM"):
                    parser = bpdb.PDBParser()
                else:
                    parser = bpdb.MMCIFParser()

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        s = parser.get_structure('temp', in_filename)
    if len(s) > 1:
        warnings.warn("Multiple models in file. Using only the first model")

    # Let's detach all H2O, to speed up processing.
    for chain in s[0]:
        log.debug("Before detaching water from %s: chain has %s residues", chain.id, len(chain))
        for r in chain.child_list[:]: # We need a copy here, because we are modifying it during iteration
            if r.resname.strip() == "HOH":
                chain.detach_child(r.id)
        log.debug("After detaching water from %s: chain has %s residues", chain.id, len(s[0][chain.id]))

    # Rename residues from other programs
    for chain in s[0]:
        for r in chain:
            # rename rosetta-generated structures
            # Old: with whitespace
            if r.resname == ' rA':
                r.resname = '  A'
            elif r.resname == ' rC':
                r.resname = '  C'
            elif r.resname == ' rG':
                r.resname = '  G'
            elif r.resname == ' rU':
                r.resname = '  U'
            # newer biopython: No whitespace
            if r.resname == 'rA':
                r.resname = 'A'
            elif r.resname == 'rC':
                r.resname = 'C'
            elif r.resname == 'rG':
                r.resname = 'G'
            elif r.resname == 'rU':
                r.resname = 'U'
            # rename iFoldRNA-generated structures
            if r.resname == 'ADE':
                r.resname = '  A'
            elif r.resname == 'CYT':
                r.resname = '  C'
            elif r.resname == 'GUA':
                r.resname = '  G'
            elif r.resname == 'URI':
                r.resname = '  U'
    # Now search for protein interactions.
    if not no_annotation:
        interacting_residues = enumerate_interactions_kdtree(s[0])
    else:
        interacting_residues = set()

    # The chains containing RNA
    chains = list(chain for chain in s[0] if contains_rna(chain))
    
    if isinstance(parser, bpdb.MMCIFParser):
        cifdict = parser._mmcif_dict # We read a private attribute here, because parsing the mmcif dictionary a second time would cause a performance penalty.
        # Generate an assembly
        try:
            operation_mat, operation_vec = _extract_symmetrymatrices_from_cif_dict(cifdict)
        except KeyError:
            pass
        else:
            if False: # Still experimental and not working correctly.
                chains = _get_assemblies(chains, cifdict)
        mr = []
        try:
            mask = np.array(
                cifdict["_pdbx_poly_seq_scheme.pdb_mon_id"], dtype=str) == "?"
            int_seq_ids = np.array(
                cifdict["_pdbx_poly_seq_scheme.pdb_seq_num"], dtype=int)[mask]
            cs = np.array(
                cifdict["_pdbx_poly_seq_scheme.pdb_strand_id"], dtype=str)[mask]
            insertions = np.array(
                cifdict["_pdbx_poly_seq_scheme.pdb_ins_code"], dtype=str)[mask]
            insertions[insertions == "."] = " "
            symbol = np.array(
                cifdict["_pdbx_poly_seq_scheme.mon_id"], dtype=str)[mask]
        except KeyError:
            pass
        else:
            if not no_annotation:
                for i, sseq in enumerate(int_seq_ids):
                    mr.append({
                        "model": None,
                        "res_name": symbol[i],
                        "chain": cs[i],
                        "ssseq": sseq,
                        "insertion": insertions[i]
                    })

    else:
        log.debug("PDB header %s", parser.header)
        try:
            mr = parser.header["missing_residues"]
            if assembly_nr is not None:
                warnings.warn("Getting an assembly is not supported for the old PDB format.")
        except KeyError:
            mr = []
            with open(in_filename) as f:
                for wholeline in f:
                    if wholeline.startswith("REMARK 465"):
                        line = wholeline[10:].strip()
                        mr_info = _parse_remark_465(line)
                        if mr_info is not None:
                            mr.append(mr_info)
                    else:
                        continue

    if mr:
        log.info("This PDB has missing residues")
    elif not no_annotation:
        log.info("This PDB has no missing residues")
    log.debug("LOADING DONE: chains %s, mr %s, ir: %s",
             chains, mr, interacting_residues)
    return chains, mr, interacting_residues



def _parse_remark_465(line):
    """Parse missing residue remarks.
    Returns a dictionary describing the missing residue.
    The specification for REMARK 465 at
    http://www.wwpdb.org/documentation/file-format-content/format33/remarks2.html#REMARK%20465
    only gives templates, but does not say they have to be followed.
    So we assume that not all pdb-files with a REMARK 465 can be understood.
    Returns a dictionary with the following keys:
    "model", "res_name", "chain", "ssseq", "insertion"
    """
    if line:
        # Note that line has been stripped.
        assert line[0] != " " and line[-1] not in "\n ", "line has to be stripped"
    pattern = (r"""
                (\d+\s[\sA-Z][\sA-Z][A-Z] |   # Either model number + residue name
                 [A-Z]?[A-Z]?[A-Z])           # Or only residue name with
                                              # 1 (RNA) to 3 letters
                \s ([A-Za-z0-9])              # A single character chain
                \s+(\d+[A-Za-z]?)$            # Residue number: A digit followed
                                              # by an optional insertion code
                                              # (Hetero-flags make no sense in
                                              # context with missing res)
                """)
    match = re.match(pattern, line, re.VERBOSE)
    if match is not None:
        residue = {}
        if " " in match.group(1):
            model, residue["res_name"] = match.group(1).split(" ")
            residue["model"] = int(model)
        else:
            residue["model"] = None
            residue["res_name"] = match.group(1)
        residue["chain"] = match.group(2)
        try:
            residue["ssseq"] = int(match.group(3))
        except ValueError:
            residue["insertion"] = match.group(3)[-1]
            residue["ssseq"] = int(match.group(3)[:-1])
        else:
            residue["insertion"] = None
        return residue
    return None

def enumerate_interactions_kdtree(model):
    relevant_atoms = [a for a in model.get_atoms() if a.name[0] in [
        "C", "N", "O"]]
    if not relevant_atoms:
        return set()
    kdtree = bpdb.NeighborSearch(relevant_atoms)
    pairs = kdtree.search_all(6, "A")
    res_pair_list = set()
    for a1, a2 in pairs:
        if a1.name not in all_side_chains and a2.name not in all_side_chains:
            continue
        p1 = a1.get_parent()
        p2 = a2.get_parent()
        if p1.id == p2.id:
            continue
        elif p1 < p2:
            res_pair_list.add((p1, p2))
        else:
            res_pair_list.add((p2, p1))
    interacting_residues = set()
    for res1, res2 in res_pair_list:
        rna_res = None
        other_res = None
        if res1.resname.strip() in RNA_RESIDUES and not res1.id[0].startswith("H_"):
            rna_res = res1
        else:
            other_res = res1
        if res2.resname.strip() in RNA_RESIDUES and not res2.id[0].startswith("H_"):
            rna_res = res2
        else:
            other_res = res2
        if rna_res is None or other_res is None:
            continue
        log.debug("%s(chain %s) and %s(chain %s, resname %s) are close", rna_res,rna_res.parent.id,other_res, other_res.parent.id, other_res.resname.strip())
        # Only consider C and N. So no ions etc
        if any(a.name[0] in ["C", "N"] for a in other_res.get_atoms()):
            interacting_residues.add(rna_res)
        else:
            log.debug("but %s has wrong atoms %s", other_res,
                      list(a.name for a in other_res.get_atoms()))
    log.debug("Interacting: %s", interacting_residues)
    return interacting_residues


"""def residues_interact(rna_res, other_res):
    for rna_atom in rna_res:
        if rna_atom.get_name() in all_side_chains:
            for other_atom in other_res:
                atom_symbol="".join(s for s in other_atom.get_name() if not s.isdigit())
                if atom_symbol in ["C", "N"]:
                    d=ftuv.vec_distance(rna_atom.coord, other_atom.coord)
                    if d<6:
                        return True
    return False"""

HBOND_CUTOFF = 4.5  # 4.5 and 0.9 are values optimized against DSSR for 5T5H_A-B-C
OOP_CUTOFF = 0.9


def _get_points(res1, res2):
    labels = {res1.resname.strip(), res2.resname.strip()}
    if labels == {"A", "U"}:
        return _points_AU(res1, res2)
    elif labels == {"G", "C"}:
        return _points_GC(res1, res2)
    elif labels == {"G", "U"}:
        return _points_GC(res1, res2)
    else:
        return None


def _points_AU(res1, res2):
    if res1.resname.strip() == "A":
        resA = res1
        resU = res2
    else:
        resA = res2
        resU = res1
    a = resA["N6"].coord
    b = resU["O4"].coord
    c = resA["N1"].coord
    d = resU["N3"].coord
    return (resA["C8"].coord, resU["C6"].coord), (a, b), (c, d)


def _points_GU(res1, res2):
    if res1.resname.strip() == "G":
        resG = res1
        resU = res2
    else:
        resG = res2
        resU = res1
    a = resG["O6"].coord
    b = resU["N3"].coord
    c = resG["N1"].coord
    d = resU["O2"].coord
    return (resG["C8"].coord, resU["C6"].coord), (a, b), (c, d)


def _points_GC(res1, res2):
    if res1.resname.strip() == "G":
        resG = res1
        resC = res2
    else:
        resG = res2
        resC = res1
    a = resG["O6"].coord
    c = resG["N1"].coord
    e = resG["N2"].coord
    b = resC["N4"].coord
    d = resC["N3"].coord
    f = resC["O2"].coord
    return (resC["C6"].coord, resG["C8"].coord), (a, b), (c, d), (e, f)


def is_basepair_pair(res1, res2):

    pairs = _get_points(res1, res2)
    if not pairs:
        return False
    for pair in pairs[1:]:  # pairs[0] is only for coplanarity]]
        d = ftuv.vec_distance(pair[0], pair[1])
        if d >= HBOND_CUTOFF:
            return False
    if is_almost_coplanar(*[point for pair in pairs for point in pair]):
        return True
    return False


def _coplanar_point_indices(*points):
    """ Thanks to https://stackoverflow.com/a/18968498"""
    from numpy.linalg import svd
    points = np.array(points).T
    assert points.shape[0] <= points.shape[1], "There are only {} points in {} dimensions.".format(
        points.shape[1], points.shape[0])
    ctr = points.mean(axis=1)
    x = points - ctr[:, np.newaxis]
    M = np.dot(x, x.T)  # Could also use np.cov(x) here.
    normal = svd(M)[0][:, -1]

    out = []
    for i, p in enumerate(points.T):
        w = p - ctr
        oop_distance = ftuv.magnitude(
            np.dot(w, normal)) / ftuv.magnitude(normal)
        if oop_distance <= OOP_CUTOFF:
            out.append(i)
    return out, ctr, normal


def is_almost_coplanar(*points):
    indices, c, n = _coplanar_point_indices(*points)
    return len(indices) == len(points)


def annotate_fallback(chain_list):
    """
    If neither DSSR nor MC-Annotate are available, we use an ad-hoc implementation of canonical
    basepair detection as fallback.
    This does not work well for missing atoms or modified residues.
    """
    kdtree = bpdb.NeighborSearch(
        [atom for chain in chain_list for atom in chain.get_atoms()])
    pairs = kdtree.search_all(10, "R")
    basepairs = {}
    # Sorted, so conflicting basepairs are deterministically solved
    for res1, res2 in sorted(pairs):
        if res1.resname.strip() not in RNA_RESIDUES or res1.id[0].startswith("H_"):
            continue
        if res2.resname.strip() not in RNA_RESIDUES or res2.id[0].startswith("H_"):
            continue
        labels = {res1.resname.strip(), res2.resname.strip()}
        try:
            is_bp = is_basepair_pair(res1, res2)
            if is_bp:
                res1_id = fgr.resid_from_biopython(res1)
                res2_id = fgr.resid_from_biopython(res2)
                if res1_id in basepairs:
                    warnings.warn("More than one basepair detected for {}."
                                  " Ignoring {}-{} because {}-{} is already"
                                  " part of the structure".format(res1_id, res1_id, res2_id, res1_id, basepairs[res1_id]))
                    continue
                if res2_id in basepairs:
                    warnings.warn("More than one basepair detected for {}."
                                  " Ignoring {}-{} because {}-{} is already"
                                  " part of the structure".format(res2_id, res2_id, res1_id, res2_id, basepairs[res2_id]))
                    continue
                basepairs[res1_id] = res2_id
                basepairs[res2_id] = res1_id
        except KeyError as e:
            log.debug("Missing atom %s. %s has atoms %s, %s has atoms %s",
                      e, res1, res1.child_dict, res2, res2.child_dict)
            pass

    seq_ids = []
    for chain in sorted(chain_list, key=lambda x: x.id):
        for residue in chain:
            seq_ids.append(fgr.resid_from_biopython(residue))
    bpseq = ""
    chain_dict = {c.id: c for c in chain_list}
    for i, seqid in enumerate(seq_ids):
        if seqid in basepairs:
            bp = seq_ids.index(basepairs[seqid]) + 1
        else:
            bp = 0

        bpseq += "{} {} {}\n".format(i + 1,
                                     chain_dict[seqid.chain][seqid.resid].resname.strip(
                                     ),
                                     bp)
    return bpseq, seq_ids


def rename_rosetta_atoms(chain):
    '''
    Rosetta names all the backbone atoms with an asterisk rather than an
    apostrophe. All that needs to be reversed.

    :param chain. A Bio.PDB.Chain structure generated by Rosetta
    :return: The same chain with renamed atoms
    '''
    for a in bpdb.Selection.unfold_entities(chain, 'A'):
        oldid = a.id
        a.name = a.name.replace('*', "'")
        a.fullname = a.name.replace('*', "'")
        a.id = a.id.replace('*', "'")
        #: Not needed with newer biopython versions
        #: Seems to be needed again?
        del a.parent.child_dict[oldid]
        a.parent.child_dict[a.id] = a
    # log.debug("Replaced rosetta atoms. \n%s\n%s",
    #                    chain.child_list[0].child_list,
    #                    chain.child_list[0].child_dict
    #                    )

    return chain


def remove_disordered(chain):
    for i, residue in enumerate(chain):
        if hasattr(residue, "selected_child"):
            new_res = residue.selected_child
            chain.detach_child(residue.id)
            chain.insert(i, new_res)
            residue = new_res
        for j, atom in enumerate(residue):
            if hasattr(atom, "selected_child"):
                new_atom = atom.selected_child
                new_atom.altloc = " "
                new_atom.occupancy = 1.0
                new_atom.disordered_flag = 0
                residue.detach_child(atom.id)
                residue.insert(j, new_atom)
    return chain


def load_structure(pdb_filename):
    '''
    Load a Bio.PDB.Structure object and return the largest chain.
    This chain will be modified so that all hetatms are removed, modified
    residues will be renamed to regular residues, etc...

    '''
    chain, mr, ir = get_biggest_chain(pdb_filename)
    return clean_chain(chain, True)[0]


def clean_chain(chain, query_PDBeChem=False):
    """
    Clean a pdb chain for further use with forgi.

    It will be modified so that all hetatms are removed, modified
    residues will be renamed to regular residues, residue ids will be positive integers, ...

    :param chain: A Bio.PDB.Chain object
    :param query_PDBeChem: If true, query the PDBeChem database whenever a
                          modified residue with unknown 3-letter code
                          is encountered.

    :returns: A modified version of this chain
    """
    chain, modifications = to_4_letter_alphabeth(chain, query_PDBeChem)
    chain = rename_rosetta_atoms(chain)
    chain = remove_disordered(chain)
    return chain, modifications


def interchain_contacts(struct):
    all_atoms = bpdb.Selection.unfold_entities(struct, 'A')

    ns = bpdb.NeighborSearch(all_atoms)
    pairs = ns.search_all(2.8)

    ic_pairs = []

    for (a1, a2) in pairs:
        if a1.parent.parent != a2.parent.parent:
            ic_pairs += [(a1, a2)]
    return ic_pairs


def contains_rna(chain):
    '''
    Determine if a Bio.PDB.Chain structure corresponds to an RNA
    molecule.

    :param chain: A Bio.PDB.Chain molecule
    :return: True if it is an RNA molecule, False if at least one residue is not an RNA.
    '''
    for res in chain:
        if res.resname.strip() in RNA_RESIDUES:
            return True
    return False


def is_protein(chain):
    '''
    Determine if a Bio.PDB.Chain structure corresponds to an protein
    molecule.

    :param chain: A Bio.PDB.Chain molecule
    :return: True if it is a protein molecule, False otherwise
    '''
    for res in chain:
        if res.resname in ['ALA', 'ARG', 'ASN', 'ASP', 'CYS', 'GLN', 'GLU', 'HIS', 'ILE', 'LEU', 'LYS', 'MET', 'PHE', 'PRO', 'SER', 'THR', 'TRP', 'TYR', 'VAL']:
            return True
    return False
